//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      This file.                                                       *
//* CLASS    Clans.java. (Model of something)                                 *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     12/03/2018                                                       *
//* REVISION X.0.1.4.1                                                        *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  To do something.                                                 *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in Java to model proto-humans.                           *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//* TODO     create a GUI drop-down with the variables for setup              *
//* TODO     ensure IF YOU WANT TO ADD P-Type intelligence that hooks exist   *
//* TODO     create a set of three BMPs representing chimpanzee refuges       *
//*          Gombi, Tanzania; Marahoue, Ivory Coast; and the original maps    *
//*                                                                           *
//* COMPILE  javac -g -verbose -deprecation Model.java                        *
//* RUN AS   java -Xms15000m -Xmx15000m -Xss8m -jar Clans.jar -cp             *
//*                                                                           *
//*****************************************************************************
//

package root;                                         // sub-directory storage.

import java.awt.Point;                                        // Java language.
import java.sql.Date;                                         //         ditto.
import java.sql.Time;                                         //         ditto.
import java.util.ArrayList;                                   //         ditto.
import java.util.Random;                                      //         ditto.

import root.hominini.Hominid;                          // program dependencies.
import root.hominini.SeedGroup;                        //                ditto.
import root.ecology.Ecology;                           //                ditto.
import root.events.LifeExperiences;                    //                ditto.
import root.gui.GUI;                                   //                ditto.
import root.rng.TrueRNG;                               //                ditto.
import root.tools.Setup;                               //     (not used) ditto.
import root.writers.ModelReports;                      //                ditto.

//
//*****************************************************************************
// Clans

class Clans extends ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  // see ModelConstaints for system constants.

  // Variables.
  static boolean USE_FIXED = true;              // RNG seed is fixed or random.
        // if USE_FIXED=false, TrueRNG is reqd for max entropy else is factory.
     // if USE_FIXED=true, TrueRNG not reqd. Entropy is factory and calib_seed.
  // see entry in ModelConstraints.java also.

  static long RNG_SEED = USE_FIXED ? calib_seed : random_seed;  // starts RNGs.

  private static int epoch = 0;                      // the current epoch date.

  private static int agentsDead = 0;     // agent count having died in the sim.
  private static int agentsAlive = 0;         // agents currently alive in sim.
  private static int agentBirths = 0;    // total count of agent births in sim.
  private static int totalAgentsSimulated = 0;       // total agents simulated.
  private static int infantDeaths = 0;   // total count of agent deaths in sim.

  private static int mapWidth = 0;    // the horizontal size of the user's map.
  private static int mapHeight = 0;     // the vertical size of the user's map.
  private static long start = 0L;      // used for benchmarking with timeStamp.

  // Objects.
  public static GUI userGUI = null;                 // user graphics interface.
  public static Ecology ecoSys = null;      // the virtual ecosystem simulated.
  public static SeedGroup seed = null;  // used to start SeedGroup of hominids.
  public static ArrayList <Hominid> population = null;  // homis in the ecoSys.
  public static LifeExperiences life = null;    // the daily lives of hominids.
  public static ModelReports dailyReport = null;           // disk file writer.
  public static ModelReports reaperReport = null;          // disk file writer.
  public static ModelReports statusReport = null;          // disk file writer.
  public static Setup setup = null; // utility Class of program specific tools.
                                                      // not fully implemented.

  public static Random RNG = new Random(RNG_SEED);// a random number generator.
  public static TrueRNG trueRNG = new TrueRNG(true_rng_port);   // TrueRNG USB.

  //***************************************************************************
  /** This method reports: hominids still alive, births and deaths recorded,
      and the total count of agents simulated (including currently) so far.*/

  public static void agentsInSimReport (int epoch)
  {

    int percentInfantMortality = (int) (((float) infantDeaths /
                                         (float) agentBirths) * 100.0F);
    int avgCSize = ecoSys.averageCohortSize ();
    int avgBirthRate = (epoch <= YEAR ? agentBirths :
                       ( agentBirths / ( (int) (epoch/YEAR) )));
    int productivity = (int) ecoSys.terrainProductivity(50.0F);   // all > 50%;
    int incestCount = life.getIncestCount();

    System.out.println("####################################################" +
                       "####################################################");
    System.out.println("Clans.agentsInSimReport()");
    System.out.println(

    "Year nr: ................... " + (epoch/YEAR) + "\n" +
    "Day nr: .................... " + ((epoch % YEAR) + 1) + "\n" +
    "Terrain productivity: ...... " + productivity + "%\n" +
    "Hominini still alive: ...... " + agentsAlive + "\n" +
    "Average cell occupancy: .... " + (avgCSize == -1?"0": avgCSize) + "\n" +
    "Hominini deaths recorded: .. " + agentsDead + "\n" +
    "Hominini births recorded: .. " + agentBirths + "\n" +
    "Avg birth rate (yearly): ... " + avgBirthRate + "\n" +
    "Percent infant mortality: .. " + percentInfantMortality + "\n" +
    "Incest count recorded: ..... " + incestCount + "\n" +
    "Total simulated so far: .... " + totalAgentsSimulated + "\n");

    System.out.println("####################################################" +
                       "####################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the closing announcements.*/

  public static void closingAnnouncements ()
  {

    int years = (int) (epoch / YEAR);
    int days = epoch - (years * YEAR);

    System.out.println("FINAL:\n");
    agentsInSimReport(epoch);                   // DIAGNOSTIC - can be removed.

    // ------------------------------------------------------------------------
    System.out.println("ANALYSIS:\n");
    System.out.println("Simulated interval was " + years + " years " +
                       days + " days.");
    System.out.println("Real-time: " + timeStamp (start));
    System.out.println("Average time per sim-year: " +
                      ((float) epoch / (float) 364) + " sec/year\n");
    // ------------------------------------------------------------------------

    start = 0;                                        // benchmarking timehack.
    System.out.println("#################################################\n");
    System.out.println("The random number seed " +
                     (USE_FIXED ?
                     "used was: " + RNG_SEED : "generated by TrueRNG") + ".");
    reportDateAndTime();                     // calendar date and current time.
    System.out.println("Halted and waiting.");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs methods in ModelReports in order to construct a set
      of coherent reports of any files in the data folder.*/

  public static void collateReports ()
  {

    final int REAPER = 0; // symbol for the Reaper Report path in ModelReports.
    final int DAILY = 1;   // symbol for the Daily Report path in ModelReports.

    System.out.println("Data folder file collation requested.");

    // here we first construct the REAPER report then create its offspring.
    statusReport.collateReports(REAPER);    // bind many Reaper files into one.
    statusReport.collateSocialNets();     // bind the geneological social nets.
    statusReport.codHistory();             // bind cause of death history data.
    statusReport.deathAnnuals();             // bind annual death history data.
    statusReport.pairingsReport();         // bind mating and breeding reports.

    // here we next construct the DAILY report then create its offspring.
    statusReport.collateReports(DAILY);      // bind many Daily files into one.
    statusReport.collateYearlyStats();                    // bind yearly notes.
    statusReport.seasonalProductivity();      // bind terrain stats seasonally.

    System.out.println("Data folder collations completed.");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports: hominids still alive, births and deaths recorded,
      and the total count of agents simulated (including currently) so far.*/

  public static String dailyStatistics (int epoch)
  {

    int percentInfantMortality = (int) (((float) infantDeaths /
                                         (float) agentBirths) * 100.0F);
    int avgCSize = ecoSys.averageCohortSize ();
        avgCSize = (avgCSize == -1 ? 0: avgCSize);

    int avgBirthRate = (epoch <= YEAR ? agentBirths :
                       ( agentBirths / ( (int) (epoch/YEAR) )));
    int productivity = (int) ecoSys.terrainProductivity(50.0F);
    int incestCount = life.getIncestCount();

    String s = new String (Integer.toString (epoch) + "," +
                           Integer.toString (epoch/YEAR) + "," +
                           Integer.toString ((epoch % YEAR) + 1) + "," +
                           Integer.toString (productivity) + "," +
                           Integer.toString (agentsAlive) + "," +
                           Integer.toString (avgCSize) + "," +
                           Integer.toString (agentsDead) + "," +
                           Integer.toString (agentBirths) + "," +
                           Integer.toString (avgBirthRate) + "," +
                           Integer.toString (percentInfantMortality) + "," +
                           Integer.toString (incestCount) + "," +
                           Integer.toString (totalAgentsSimulated) + "\r\n");
    return s;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports: the random number seed used, date of birth,
      date of death, chromosome, Agent ID, Father ID, Mother ID, x birth
      location, y birth location, x death location, y death location, age,
      gender, pregnancy status, current weight, caloric requirement, water
      requirement, life status, alpha status, and cause of death.  The
      output in a comma delimited string.*/

  public static String deathCertificate (Hominid h)
  {

    String s = new String ( RNG_SEED + "," +
               epoch + "," +
               h.getBirthDate() + "," +
               h.getDeathDate() + "," +
               h.aGenes.getChromosomeString(h.bio.getChromosome()) + "," +
               h.getAgentID() + "," +
               h.getMotherID() + "," +
               h.getFatherID() + "," +
               h.getBirthLoc().x + "," +
               h.getBirthLoc().y + "," +
               h.getDeathLoc().x + "," +
               h.getDeathLoc().y + "," +
               h.bio.getAge() + "," +
               h.getSexReport() + "," +
               h.bio.getPregnancyStatus() + "," +
               h.bio.getCurrentWeight() + "," +
               h.bio.getCaloricRequirement() + "," +
               h.bio.getWaterRequirement() + "," +
               h.kin.getIsAlpha() + "," +
               h.getPreferredPrey() + "," +
               h.getCauseOfDeath() + "\n" );

    return s;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method serves as a DIAGNOSTIC utility.  Several varied tools may
      reside here.*/

  public static void diagnosticUtilities ()
  {

    // driveAnalysis (population);         // displays population drive trends.

    //if (epoch > (4 * YEAR))
    //if (epoch > 5800 && epoch < 5810)
    //{

      for (Hominid h : population)
      {

        // group-specific tracking.
        if (h.bio.getHominidLifeStatus().equals("J"))
          h.reportMetabolism(); // metabolism specific.

        // age specific tracking.
        //if (h.bio.getAge() > 5 && h.bio.getAge() <= 12)
        //{
        //    h.reportMetabolism(); // metabolism specific.
        //}

      }

    //}

    //if (epoch == 5811) System.exit(0);


  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the instantaneous totals of the hominids drives.*/

  public static ArrayList <Integer> driveAnalysis
               (ArrayList <Hominid> population)
  {

    String sMask = "00000000";
    int panicTotal = 0;
    int thirstTotal = 0;
    int hungerTotal = 0;
    int groupTotal = 0;
    int dispersalTotal = 0;
    int na_1Total = 0;
    int na_2Total = 0;
    int thermalTotal = 0;

    for (Hominid h : population)
    {
      sMask = h.getActiveDrives ();
      panicTotal     += (sMask.substring(0, 1).equals("1") ? 1 : 0);
      thirstTotal    += (sMask.substring(1, 2).equals("1") ? 1 : 0);
      hungerTotal    += (sMask.substring(2, 3).equals("1") ? 1 : 0);
      groupTotal     += (sMask.substring(3, 4).equals("1") ? 1 : 0);
      dispersalTotal += (sMask.substring(4, 5).equals("1") ? 1 : 0);
      na_1Total      += (sMask.substring(5, 6).equals("1") ? 1 : 0);
      na_2Total      += (sMask.substring(6, 7).equals("1") ? 1 : 0);
      thermalTotal   += (sMask.substring(7, 8).equals("1") ? 1 : 0);
    }

    ArrayList <Integer> list = new ArrayList <Integer> ();
    list.add (new Integer(panicTotal));
    list.add (new Integer(thirstTotal));
    list.add (new Integer(hungerTotal));
    list.add (new Integer(groupTotal));
    list.add (new Integer(dispersalTotal));
    list.add (new Integer(na_1Total));
    list.add (new Integer(na_2Total));
    list.add (new Integer(thermalTotal));

    System.out.println("####################################################");
    System.out.println("Clans.driveAnalysis() yearly drive priorities.\n");
    System.out.println (" hasPanic:         " + panicTotal);
    System.out.println (" hasThirst:        " + thirstTotal);
    System.out.println (" hasHunger:        " + hungerTotal);
    System.out.println (" needsSocialGroup: " + groupTotal);
    System.out.println (" needsDispersal:   " + dispersalTotal);
    System.out.println (" undefined:        " + na_1Total);
    System.out.println (" undefined:        " + na_1Total);
    System.out.println (" hasDiscomfort:    " + thermalTotal);
    System.out.println("####################################################");

    return list;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the elapsed time, actual date and time.*/

  public static void reportDateAndTime ()
  {

    Date currentDate = new Date(System.currentTimeMillis());
    Time currentTime = new Time(System.currentTimeMillis());
    System.out.println("Calendar date is: " + currentDate +
                       " / Actual time is: " + currentTime + "\n");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the simulation.*/

  public static void setupEnvironment (int mapWidth, int mapHeight)
  {

    // report the date and time this simulation began.
    System.out.println("\n\n" + "Begin full system initialization.");
    System.out.println("________________________________________________" +
    "\n");

    if (USE_FIXED)
    {
      System.out.println("TrueRNG USB not selected.");
      System.out.println("The random number seed used was: " + RNG_SEED + ".");
    }
    else
    {
      if (!trueRNG.setupDevice())
      {
        System.out.println("TrueRNG USB setup failed.");
        System.exit(0);
      }
      trueRNG.startDevice ();
    }

    epoch = 0;                                       // the current epoch date.
    agentsDead = 0;                     // accounting of agent data in the sim.
    agentsAlive = 0;                                                  // ditto.
    agentBirths = 0;                                                  // ditto.
    totalAgentsSimulated = 0;                                         // ditto.
    infantDeaths = 0;                                                 // ditto.

    ModelReports folder = new ModelReports ();   // clear data folder contents.
    folder.clearFilesNamed ("txt");      // delete all files with .txt in name.
    folder = null;                // return the memory to the operating system.

    dailyReport = new ModelReports ("DailyReport.txt");   // save data to disk.
    dailyReport.open(epoch);                           // open the output file.
    statusReport = new ModelReports ();       // manipulate saved data reports.
    reaperReport = new ModelReports ("ReaperReport.txt"); // save data to disk.
    reaperReport.open(epoch);                          // open the output file.

    // this is the synthetic ecology within which the Sim occurs.
    ecoSys = new Ecology (RNG_SEED, (USE_FIXED ? null : trueRNG),
                          mapWidth, mapHeight);      // setup the terrain grid.
    userGUI.getHeightMap (ecoSys); // use map colors as elevations for Ecology.
    ecoSys.initializeEcology ();  // based on HghtMap setup upr&lwr ecoSystems.

    // these are the first Hominids in the simulation.
    population = new ArrayList <Hominid>();

    // these will be the daily lives of the hominids.
    life = new LifeExperiences (
        (USE_FIXED ? RNG_SEED : trueRNG.getNextRandom()), mapWidth, mapHeight);
    life.setIncestCount(0);             // accounting of agent data in the sim.

    // report status of experimental variables.
    System.out.println (use_altruism?"Altruism":"Promiscuity" + " in force.");
    String gs = Integer.toString(maximum_group_size);       // seed-group size.
    String ng = Integer.toString(number_of_groups);       // nr of seed-groups.
    System.out.println ( ng + " X " + gs + " " +
                       ( fixed_grid ? "fixed grid" : "random grid." ) );
    System.out.println ();

    // a 6-step sequence to start-up, or "seed," the hominids on the grid.
    seed = new SeedGroup (RNG_SEED, (USE_FIXED ? null : trueRNG),
                                     mapWidth, mapHeight, life);
    population.addAll (seed.gridStart (epoch,
                                       maximum_group_size,
                                       number_of_groups,
                                       life));
    agentsAlive = seed.getAgentsAlive (); // start bookkeeping on agent totals.
    totalAgentsSimulated = seed.getTotalAgentsSimulated ();           // ditto.
    ecoSys.initializeGridCohorts (population);
    seed = null;                  // return the memory to the operating system.

    System.out.println("________________________________________________");
    System.out.println("\nFull system initialization complete.");

    System.gc();        // in case any garbage was made during the above steps.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method shuts down the simulation (data files).*/

  public static void shutdownEnvironment ()
  {

    if (!USE_FIXED)
    {
      trueRNG.stopDevice();                    // close the TrueRNG USB device.
    }

      reaperReport.close ();     // close the buffer and transfer data to disk.
      dailyReport.close ();      // close the buffer and transfer data to disk.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method reports the starting announcements.*/

  public static void startingAnnouncements ()
  {

    System.out.println("________________________________________________");
    System.out.println("\nSimulation starting.");           // announce status.
    reportDateAndTime();                     // calendar date and current time.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method removes Hominid objects from the simulation.  Additionally,
      it provides an optional output report for those agent(s) processed.
      The process involves scanning each person in the list scanning for
      those agents marked as "D" dead.  These agents are removed and the
      population shortened and returned to the caller.*/

  public static ArrayList <Hominid> theGrimReaper (int epoch,
                                                  ArrayList <Hominid> homis)
  {

    Hominid coffin = null;      // you may not need it, but get a coffin ready.
    for (int i = (homis.size() - 1); i >= 0; i--)
    {

      if ( homis.get(i).bio.getHominidLifeStatus().equals("D") )
      {

        coffin = homis.get(i);             // put the individual in the coffin.
        Point deathLocation = coffin.getCurrentLoc();    // get place of death.
        coffin.setDeathLoc(deathLocation);            // record place of death.
        String cod = coffin.getDeathReport();        // get the cause Of death.
        coffin.setCauseOfDeath(cod);              // record the cause Of death.
        coffin.setDeathDate(epoch);                // record the date of death.
        String lifePhase = coffin.bio.getLifeGroup ();  // life state at death.

        //identify which grid cell the coffin came from.
        int h = coffin.getCurrentLoc().x;             // this is the cell in X.
        int v = coffin.getCurrentLoc().y;             // this is the cell in Y.

        // remove this hominid from the cell's cohort list and clear memory.
        ecoSys.removeCohortIndividual (h, v, coffin);           // out of cell.
        homis.remove (coffin);  // dispose of the remains (release the memory).

        //coffin.report(epoch);   // report the post-mortem results to display.
        reaperReport.accumulate (deathCertificate(coffin)); // records to disk.

        // --------------------------------------------------------------------
        // update all clerical accounts associated with the death of an agent.
        agentsDead++;                               // increment "dead" agents.
        agentsAlive--;                            // decrement "living" agents.
        infantDeaths += (lifePhase.equals ("B") ? 1 : 0);  // if (B) increment.
        // --------------------------------------------------------------------

      }// end if.

    }// end for loop.

    coffin = null;                // return the memory to the operating system.

    return homis;            // return the remaining population to circulation.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is reponsible for adding all new Hominid objects to the
      population ArrayList.  Effectively, this is where birth happens.*/

  public static ArrayList <Hominid> theStork (int epoch,
                                              ArrayList <Hominid> homis)
  {

    Hominid hominid = null;                         // a potential hominid mom.
    for (int i = 0; i < homis.size(); i++)
    {

      hominid = homis.get(i);                          // examine this hominid.

      // if female and ready to give birth.
      if (hominid.bio.getSex().equals("F") && hominid.bio.getNewBirthNotice())
      {

        hominid.bio.setNewBirthNotice(false); // if so, reset the birth notice.
        int tempCount = hominid.bio.getLifetimeBirths();
        hominid.bio.setLifetimeBirths(++tempCount);    // update running total.

        Hominid father = hominid.bio.getMObject();    // retrieve father to be.

        // instantiate a new, baby hominid.
        Hominid child = new Hominid (RNG_SEED, (USE_FIXED ? null : trueRNG),
                            hominid, father, epoch, mapWidth, mapHeight, life);
        homis.add ( child );                 // add the baby to the population.

        // mom and child share apriori kin pointers.   Dad doesn't get pntr.
        hominid.kin.addChild(child); // give mom pntr to child - MUST DO FIRST.
        hominid.kin.informSiblingsOf (child);// tell other kids about this one.
        child.kin.setMother (hominid);            // give child pointer to mom.
        child.kin.updateGroup (hominid);  // put mom into child's group memory.
        hominid.bio.setMObject (null);   //clear dad's genetic marker from mom.

        //add the child to the cell list.
        int h = hominid.getCurrentLoc().x;            // this is the cell in X.
        int v = hominid.getCurrentLoc().y;            // this is the cell in Y.
        ecoSys.addCohortIndividual (h, v, child);          // add to cell list.

        // DIAGNOSTIC
        //System.out.println("This was a birth!  Today's date is: " + epoch);
        //hominid.report(); // mothers report
        //father.report(); // fathers report
        //child.report(); // babys report
        // DIAGNOSTIC

        // --------------------------------------------------------------------
        // update all clerical accounts associated with addition of an agent.
        totalAgentsSimulated++;                  // increment the total number.
        agentsAlive++;                       // increment newly "alive" agents.
        agentBirths++;                       // increment the "new born" count.
        // --------------------------------------------------------------------

      }// end if.
    }// end for loop.

    return homis;

  } // End method.
  //***************************************************************************

  // **************************************************************************
  /** This method reports the elapsed time interval.*/

  public static String timeStamp ( long start )
  {

    final long DAY = 86400;                        // 1 day's worth of seconds.
    final long HRS = 3600L;                       // 1 hour's worth of seconds.
    final long MIN = 60L;                       // 1 minute's worth of seconds.

    long end = System.currentTimeMillis ();
    long interval = ( end - start ) / 1000L;

    long day = 0L;
    long hrs = 0L;
    long min = 0L;
    long sec = 0L;

    if ( interval >= DAY )
    {
      day = interval / DAY;
      interval -= ( day * DAY );
    }

    if ( interval >= HRS )
    {
      hrs = interval / HRS;
      interval -= ( hrs * HRS );
    }

    if ( interval >= MIN )
    {
      min = interval / MIN;
      interval -= ( min * MIN );
    }

    sec = interval;

    String elapsedTime = "@ " + day + " days " + hrs + " hrs. " +
                                min + " min. " + sec + " sec. " +
                                "(" + ( ( end - start ) / 1000L ) + ")";

    //System.out.println ( "@ " + day + " days " + hrs + " hrs. " +
    //                            min + " min. " + sec + " sec. " +
    //                            "(" + ( ( end - start ) / 1000L ) + ")" );

    return elapsedTime;

  } // End method.
  // **************************************************************************

  //***************************************************************************
  /** This is the main method.  It calls everything into existence and controls
      the cyclic activation of all simulation components and activities. Each
      of the respective control sections are clear demarcated.*/

  public static void main ( String [] arg )
  {

    setup = new Setup();          // utility Class with program specific tools.
    USE_FIXED = setup.readConsole(arg);  // read the console runtime specifics.
    userGUI = new GUI(setup);            // a graphical user interface display.

    // ---------------------- LOOP UNTIL USER STARTS --------------------------
    while (!userGUI.getUserStartRequest()) {}; //loop until user wants a start.
    // ---------------------- LOOP UNTIL USER STARTS --------------------------

    setup.closeSetupForm();               // close the user choices setup form.
    setup = null;                 // return the memory to the operating system.

    do
    {
      // -------------- BASED ON USER CHOICES BUILD THE SIM -------------------
      // setup simulation using size and map information taken from user GUI.
      mapWidth = userGUI.getMapWidth();             // these are now available.
      mapHeight = userGUI.getMapHeight();                             // ditto.
      setupEnvironment( mapWidth, mapHeight );
      startingAnnouncements ();                   // say startup announcements.
      userGUI.plotTheDisplay(ecoSys);                      // plot the display.
      start = System.currentTimeMillis ();            // benchmarking timehack.
      // -------------- BASED ON USER CHOICES BUILD THE SIM -------------------

      // ----------------------------------------------------------------------
      // ----------------- MASTER SIM CONTROL LOOP - (begin) ------------------
      do // continue until all of the population has expired or told to exit.
      {

        userGUI.setSimRunning(false);            // flag used with zoom in-out.
        while (userGUI.getUserStartRequest() && (population.size() > 0))
        {
          userGUI.setSimRunning(true);           // flag used with zoom in-out.

          // ------------------ THE ENTIRE SIMULATION -------------------------
          ++epoch;

          //diagnosticUtilities ();      // population diagnostic utility pool.
          population = theStork (epoch, population);//add new homis to the sim.
          life.doSimLife (epoch, ecoSys, population);      // simulate one day.
          population = theGrimReaper (epoch, population);   // remove the dead.

          ecoSys.ageEcoSystem (epoch);             // update simulated ecology.
          if ((epoch % userGUI.getRepRate()) == 0)
          {
            userGUI.plotTheDisplay(ecoSys);                // plot the display.
          }
          // ------------------ THE ENTIRE SIMULATION -------------------------

          // -------------------------- REPORTS -------------------------------
          dailyReport.accumulate (dailyStatistics(epoch));  // a daily archive.
          if ((epoch % 364) == 0)     // system activities executed every year.
          {
            dailyReport.restart (epoch);    // save and reopen the output file.
            reaperReport.restart (epoch);   // save and reopen the output file.
            agentsInSimReport(epoch);         // DIAGNOSTIC - console scroller.
            System.gc();     // clean-up any garbage collected upto this point
          }
          // -------------------------- REPORTS -------------------------------
        }
        if (userGUI.getPrepStatistics()) collateReports (); // collate reports.

      } while (population.size() > 0 && !userGUI.getUserStopRequest ());
      // ------------------- MASTER SIM CONTROL LOOP - (end) ------------------
      // ----------------------------------------------------------------------

      // ----------------- FINAL REPORTS & OFFER TO RESTART -------------------
      userGUI.plotTheDisplay(ecoSys);                      // plot the display.
      closingAnnouncements ();                    // say closing announcements.
      shutdownEnvironment ();          // showdown all program runtime support.
      userGUI.sleep();   // prepares the GUI for either restart or termination.
      do // NOTE: on Apple-MAC machines this may overwhelm the processor!
      {
        if (userGUI.getPrepStatistics()) collateReports (); // collate reports.
      } while (!userGUI.getRestart ());
      if (userGUI.getRestart ())
        userGUI.awaken();                   // resets GUI and refreshes screen.
      else
        System.out.println("Simulation has ended normally.");
      // ----------------- FINAL REPORTS & OFFER TO RESTART -------------------

    } while (!userGUI.getUserStopRequest ());

    System.exit(0);                                 // leave system completely.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method cause a momentary pause in ms in the executing thread.*/

  public static void wait (int delay)
  {

    try
    {
      Thread.sleep(delay);
    }
    catch (InterruptedException ie)
    {
      System.out.println ("ERROR: Clans.wait.");
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}
