//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      This file.                                                       *
//* CLASS    ModelConstraints.java. (Model Constraints)                       *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  TBD.                                                             *
//* METHOD   TBD.                                                             *
//* DETAILS  TBD.                                                             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root;                                         // sub-directory storage.

import java.io.File;                                          // Java language.

//
//*****************************************************************************
// ModelConstraints


public abstract class ModelConstraints
{

  //
  //***************************************************************************
  // GLOBAL constants.

  // --------------------------------------------------------------------------
  // Determinism, establish a basis for simulation reproduciblity.
  final public static boolean USE_FIXED = true; // RNG seed is fixed or random.
        // if USE_FIXED=false, TrueRNG is reqd for max entropy else is factory.
     // if USE_FIXED=true, TrueRNG not reqd. Entropy is factory and calib_seed.
  public static long calib_seed = 0;                           // a fixed seed.
  // see entry in Clans.java also.

  public static long random_seed = System.nanoTime();        // fast time seed.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // Arbitrary values but required constants.
  final public static int YEAR = 364;                        // a year of days.
  final public static float FAT_AS_CALORIE = 3500.0F;      // 1 lb. conversion.

  // equivalent to eating 100% of a 2700 kCal diet composed entirely of tubers.
  public static float daily_meals = 3.48F;           // "multiplier" best case.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static float min_water = 1.0F;      // min gals wtr ea basin will get.
  public static float random_water = 3.50F;      // some basins this much more.
  public static float rate_of_lowland_water_basins = 0.5F;   //% Wtr in Basins.
  public static float rate_of_highland_water_basins = 0.3F;  //% Wtr in Basins.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static float flora_success_bias = 0.50F;     // min season foragb prd.
  public static float fauna_success_bias = 0.50F;     // min season foragb prd.
  public static float initial_productivity = 99.0F; // initial percent product.
  public static float foraging_cost = 1.0F; // percent cost to terr if foraged.
  public static float foraging_rcvry_rate = 0.50F; // percent terr recvry rate.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static boolean fixed_grid = true;      // seed-groups on a fixed grid.
                                 // determines random or fixed start locations.
                                                  // choices are 1 or 2 groups.
                      // if 2 groups used then use with 400x400 or greater map.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static boolean use_altruism = true;     // engage altruistic behavior.
                                     // determines gene-based altruistic trait.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static int maximum_group_size = 100;          // max nr in seed group.
  public static int number_of_groups = 2;          // max nr of hominid groups.
                                                  // choices are 1 or 2 groups.
                      // if 2 groups used then use with 400x400 or greater map.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  final public static int DISCRETE_FACTORS_GENE =   0; // gene 0 on chromosome.
  final public static int DRAUGHT_TOLERANCE_GENE =  1; // gene 1 on chromosome.
  final public static int ROBUST_METABOLISM_GENE =  2; // gene 2 on chromosome.
  final public static int SOCIAL_ALTRUISM_GENE   =  3; // gene 3 on chromosome.
  final public static int TERRITORY_DEFENSE_GENE =  4; // gene 4 on chromosome.
  final public static int PAIR_BND_ALLIANCE_GENE =  5; // gene 5 on chromosome.
  final public static int EMIT_FERTIL_SCENT_GENE =  6; // gene 6 on chromosome.
  final public static int PHILOPAT_DISPERSE_GENE =  7; // gene 7 on chromosome.
  final public static int RESERVED_GENE_8        =  8; // gene 8 on chromosome.
  final public static int RESERVED_GENE_9        =  9; // gene 9 on chromosome.
  final public static int TEMPERTURE_SETPNT_GENE = 10; // gene A on chromosome.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static int water_days = 2;         // max consecutive epochs in water.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static int affect_word_width = 8;        // Java can support up to 30.
  public static int grouped_drives = 3;         // nr of highest priorty group.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static int fear_step = 2;          // n X 10 meter mvmnt dist per day.
  public static int water_step = 15;        // n X 10 meter mvmnt dist per day.
  public static int food_step = 15;         // n X 10 meter mvmnt dist per day.
  public static int group_step = 30;        // n X 10 meter mvmnt dist per day.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // From Biology.java (24 Jan 2015)
  // MALE: (adult maximum requirements, calculated)
  // Water = 0.750F;
  // Food  = 0.700F;
  //
  // FEMALE: (adult maximum requirements, calculated - w/reproductive loading)
  // Water = 0.975F;
  // Food  = 0.465F;

  public static float water_start_amount = 0.7F; // water amount to start with.
  public static float water_carrying_limit = 1.0F;// only hand to mouth (gals).
  public static float food_start_amount = 0.7F;    //food amount to start with.
  public static float food_carrying_limit_f = 0.7F; // only hand to mouth(lbs).
  public static float food_carrying_limit_m = 0.9F; // only hand to mouth(lbs).
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static int prey_attitude_change_threshold = 0364;     // 1 day = 0001.
                                                       //  1 week equiv = 0007.
                                                       //  1 year equiv = 0364.
                                                       //  5 year equiv = 1820.
                                                       // 10 year equiv = 3640.
                                                       // 15 year equiv = 5460.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static float chance_of_predation_hi = 0.0000140F;//dly, 0.50% yr esti.
  public static float chance_of_predation_lo = 0.0000001F;//dly, 0.04% yr Will.

  // Williams, J, et. al. (2008), Am J Prim, vol. 70, pp. 776-777. "none seen"
  // Boesch, C. (1991), Behaviour, vol. 117, nr. 3/4, pp. 220-242 (p.229-230).

  // #/y = 1 - ( 1 - p ) ^ 364; example: 0.055/y = 1 - ( 1 - 0.00016 ) ^ 364
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  public static float infanticide_mort = 0.000055F;        // dly, 2.0% annual.
  public static float disease_threshold = 0.000055F;       // dly, 2.0% annual.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // infant weaning must be complete between 4 and 6 years of age, inclusive.
  public static int range_of_early_weaning = 3;          // 0-2 by factory RNG.
  public static int min_weaning_age = 4;         // minimum oldest when weaned.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // Installation, operating system, and hardware RNG dependencies.
  final public static String BIN_FOLDER = "root" + File.separator;
  final public static String DATA_FOLDER = "data" + File.separator;
  final public static String MAP_FOLDER = "maps" + File.separator;
  public static String true_rng_port = "COM5";       // OS developed USB equiv.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // RUN_TIME must be set FALSE in order to use the two DIAGNOSTIC modes.
  public static boolean run_time_display = true;//true=normal false=DIAGNOSTIC.
  public static boolean foraging_display = true;//true=forage false=waterBasin.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // Ecological and topographic features specific to the simulation.
  public static int water = 0;          // terrestrial feature specific to Sim.
  public static int lo_land = 1;        // terrestrial feature specific to Sim.
  public static int hi_land = 2;        // terrestrial feature specific to Sim.
  // --------------------------------------------------------------------------

}
