//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Cell.java. (a single terrain cell on and in a synthetic ecology) *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  This object the bottom-level terrain cell.                       *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.ecology;                                 // sub-directory storage.

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.

import root.ecology.terrain.Terrain;                       // program routines.
import root.hominini.Hominid;                              //            ditto.

//
//*****************************************************************************
// Cell

public class Cell
{

  //
  //***************************************************************************
  // Class variables and objects.

  // Variables.
  private long rng_Seed = 0L;                    // local copy of the RNG seed.

  public int elevation = Integer.MIN_VALUE;     // absolute landmass elevation.
  public String biomeName = "";                       // land, water, ice, etc.
  public float temperature = -127.0F;   // real-valued and graded across biome.
  public float water = 0.0F;                       // available water resource.
  public float waterBaseline = 0.0F;               // available water resource.
  public String forageOption = "";  // foragable: tuber, fig, monkey, rat, TBD.
  public float forageProbability = 0.0F;  // prob foraging could be successful.
  public float maxProductivity = 0.0F;   // prctnt lvl of prdctvty at the time.

  // Objects.
  private Random RNG = null;                 // pointer to standard RNG object.

  public  ArrayList <Hominid> cohort = null; // hominini now on this grid cell.

  //***************************************************************************
  /** This is the Cell class constructor.*/

  public Cell (long seed)
  {

    this.rng_Seed = seed;                          // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);    // instantiate an RNG object with seed.

    this.elevation = 0;                         // absolute elevation or level.
    this.biomeName = "undefined";       // typically land, water, ice, or road.
    this.temperature = -127.0F;         // real-valued and graded across biome.
    this.water = 0.0F;                             // available water resource.
    this.waterBaseline = 0.0F;                     // available water resource.
    this.forageOption = "none";    // fig, leaves, monkey, termite, tuber, etc.
    this.forageProbability = 0.0F;        // prob foraging could be successful.
    this.maxProductivity = 0.0F;   // prcnt level of productivity at this time.
    this.cohort = new ArrayList <Hominid> (); // spatially co-located hominids.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}
