//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    GUI.java. (Graphic User Interface)                               *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  To provide a compact visual interface.                           *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.gui;                                            // home directory.

import java.awt.Color;                                        // Java language.
import java.awt.Component;                                    //         ditto.
import java.awt.BorderLayout;                                 //         ditto.
import java.awt.Dimension;                                    //         ditto.
import java.awt.Toolkit;                                      //         ditto.
import java.awt.event.ActionEvent;                            //         ditto.
import java.awt.event.ActionListener;                         //         ditto.
import java.awt.event.ComponentEvent;                         //         ditto.
import java.awt.event.ComponentListener;                      //         ditto.
import java.awt.event.WindowAdapter;                          //         ditto.
import java.awt.event.WindowEvent;                            //         ditto.
import java.awt.image.BufferedImage;                          //         ditto.
import java.io.File;                                          //         ditto.
import java.io.FileInputStream;                               //         ditto.
import java.io.IOException;                                   //         ditto.
import java.io.RandomAccessFile;                              //         ditto.
import javax.imageio.ImageIO;                                 //         ditto.
import javax.swing.filechooser.FileNameExtensionFilter;       //         ditto.
import javax.swing.ImageIcon;                                 //         ditto.
import javax.swing.JFileChooser;                              //         ditto.
import javax.swing.JFrame;                                    //         ditto.
import javax.swing.JLabel;                                    //         ditto.
import javax.swing.JMenu;                                     //         ditto.
import javax.swing.JMenuBar;                                  //         ditto.
import javax.swing.JMenuItem;                                 //         ditto.
import javax.swing.JOptionPane;                               //         ditto.
import javax.swing.JSeparator;                                //         ditto.
import javax.swing.JScrollPane;                               //         ditto.
import javax.swing.SwingUtilities;                            //         ditto.
import javax.swing.UIManager;                                 //         ditto.

import root.ecology.Ecology;                               // program routines.
import root.ModelConstraints;                              //            ditto.
import root.tools.Setup;                                   //            ditto.

//
//*****************************************************************************
// GUI

public class GUI  extends ModelConstraints implements ActionListener,
                                                      ComponentListener
{

  //
  //***************************************************************************
  // Class constants, properties, and objects.

  // Constants.
  final static String VERSION = "1.0.4"; // version is current revision number.

  final static String START_SCREEN_WALLPAPER = "MaelzelEmbossedLogo.gif";
  final static int  START_WINDOW_WIDTH = 640;  // a VGA video mode screen size.
  final static int  START_WINDOW_HEIGHT = 480; // a VGA video mode screen size.
  final static int  IMAGE_ZOOM_SIZE_STEP = 2;         // used with Zoom in-out.
  final static int  IMAGE_DEFAULT_SIZE = 1;                  // used with Zoom.
  final static int  IMAGE_ZOOM_MINIMUM = 1;                  // used with Zoom.
  final static int  IMAGE_ZOOM_MAXIMUM = 4;                  // used with Zoom.

  final static long DAILY_UPDATE = 1;                   // map update each day.
  final static long WEEKLY_UPDATE = 7;              // map update every 7 days.
  final static long MONTHLY_UPDATE = 28;           // map update every 28 days.
  final static long YEARLY_UPDATE = 364;          // map update every 364 days.
  final static long INITIAL_RATE = DAILY_UPDATE;    // the default update rate.

  //                             R    G    B
  final int BLACK =  new Color(  0,   0,   0).getRGB(); // the RGB black color.
  final int BLUE =   new Color(  0,   0, 255).getRGB();  // the RGB blue color.
  final int BROWN =  new Color(128,  64,   0).getRGB(); // the RGB brown color.
  final int GRAY_0 = new Color( 32,  32,  32).getRGB();   // an RGB gray color.
  final int GRAY_1 = new Color( 64,  64,  64).getRGB();   // an RGB gray color.
  final int GRAY_2 = new Color( 96,  96,  96).getRGB();   // an RGB gray color.
  final int GRAY_3 = new Color(128, 128, 128).getRGB();   // an RGB gray color.
  final int GRAY_4 = new Color(160, 160, 160).getRGB();   // an RGB gray color.
  final int GRAY_5 = new Color(192, 192, 192).getRGB();   // an RGB gray color.
  final int GRAY_6 = new Color(224, 224, 224).getRGB();   // an RGB gray color.
  final int GREEN =  new Color(  0, 255,   0).getRGB(); // the RGB green color.
  final int ORANGE = new Color(255, 128,   0).getRGB(); //the RGB orange color.
  final int PURPLE = new Color(128,   0, 128).getRGB(); //the RGB purple color.
  final int RED =    new Color(255,   0,   0).getRGB();   // the RGB red color.
  final int WHITE =  new Color(255, 255, 255).getRGB(); // the RGB white color.
  final int YELLOW = new Color(255, 255,   0).getRGB(); //the RGB yellow color.

  // Variables.
  private int horizontalLimit = 0;    // horizontal grid surface of simulation.
  private int verticalLimit = 0;        // vertical grid surface of simulation.
  private int screenHeight = 0;  // display screen dimension in the horizontal.
  private int screenWidth = 0;     // display screen dimension in the vertical.

  private long loopRate = 0;                     // the simulation update rate.
  private boolean userStartRequest = false; //flag has user sim status request.
  private boolean userStopRequest = false;  //flag has user sim status request.
  private boolean initialized = false;              // control and status flag.
  private boolean doRestart = false;                // control and status flag.
  private boolean simRunning = false;            // flag used with zoom in-out.
  public int imageZoomFactor = IMAGE_DEFAULT_SIZE;       // current image zoom.
  public boolean movementHistory = false;   // agent movement tracking control.
  private boolean prepStatisticsFiles = false;  // user data consolidation req.

  private static int bmpWidth = 0;  //derived in bmpToArray, from header 18-21.
  private static int bmpHeight = 0; //derived in bmpToArray, from header 22-25.
  private static int MS_IMAGE_TYPE = BufferedImage.TYPE_3BYTE_BGR;     //image.
  private static int BCKGND_IMAGE_TYPE = BufferedImage.TYPE_INT_ARGB;  //image.

  // GUI components and controls.
  private static JFrame jfrm;            // the program's principle GUI window.
  private static JScrollPane scrollPane;    // pointer to the map scroll panel.
  private static ImageIcon wallPaper = new ImageIcon (START_SCREEN_WALLPAPER);
  private static JLabel wPaper = new JLabel(wallPaper); // startup frame image.
  private static JMenuItem fastUpdate;       // these are drop-down menu items.
  private static JMenuItem loadMap;                                   // ditto.
  private static JMenuItem setupSim;                                  // ditto.
  private static JMenuItem prepStats;                                 // ditto.
  private static JMenuItem mediumUpdate;                              // ditto.
  private static JMenuItem quitSim;                                   // ditto.
  private static JMenuItem seedOff;                                   // ditto.
  private static JMenuItem seedOn;                                    // ditto.
  private static JMenuItem slowUpdate;                                // ditto.
  private static JMenuItem startSim;                                  // ditto.
  private static JMenuItem stopSim;                                   // ditto.
  private static JMenuItem continueSim;                               // ditto.
  private static JMenuItem stepSim;                                   // ditto.
  private static JMenuItem restartSim;                                // ditto.
  private static JMenuItem historyOff;                                // ditto.
  private static JMenuItem historyOn;                                 // ditto.
  private static JMenuItem yearUpdate;                                // ditto.
  private static JMenuItem zoomIn;                                    // ditto.
  private static JMenuItem zoomOut;                                   // ditto.
  private static JMenuItem sayHelp;                                   // ditto.
  private static JMenuItem sayVersion;                                // ditto.

  // Objects.
  private static BufferedImage bckGndImg = null;   // working, displayed image.
  private static BufferedImage msImage = null;     // default loaded map image.
  private static BufferedImage msImageScaled = null;   // scaled default image.
  private static int imageArray[] = null;  // temporary array used for loading.
  private static ImageIcon scratchPad;       // pointer to the terrain display.
  private static JLabel scratchP;                                     // ditto.

  private Setup setupForm = null;    // the user run-time variables input form.

  //***************************************************************************
  /** This inner class respond to clicks on the menu help request button. */

  private class SayHelpPane implements ActionListener
  {

    @Override
    public void actionPerformed ( ActionEvent e )
    {

      JOptionPane.showMessageDialog(jfrm, "None available at this time.");

    }
  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This inner class respond to clicks on the menu version request button.*/

  private class SayVersionPane implements ActionListener
  {

    @Override
    public void actionPerformed ( ActionEvent e )
    {

      JOptionPane.showMessageDialog(jfrm, "Version:  " + VERSION);

    }
  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This inner class respond to clicks on the reset all agents button. */

  private class Terminate implements ActionListener
  {

    @Override
    public void actionPerformed ( ActionEvent e )
    {

      int response = JOptionPane.showConfirmDialog (
        jfrm,
        "Terminate Clans?",
        "WARNING - Data May Be Lost!",
        JOptionPane.YES_NO_OPTION );

      switch ( response )
      {

        case JOptionPane.YES_OPTION:
          System.out.println ("Simulation ended normally.");
          System.exit ( 0 );
          break;                                        // choice made to quit.

        case JOptionPane.NO_OPTION:
          break;                                    // choice made to continue.

        case JOptionPane.CLOSED_OPTION:
          break;                         // closed without control choice made.

      }

    }
  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constructor instantiates an instance of the TopFormGUI object.  It
      is the class constructor and Graphical User Interface menu builder.*/

  public GUI (Setup setup)
  {

    System.out.println("\nUser map load choice pending.");  // announce status.
    System.out.println("User parameter initialization pending.");  // announce.
    System.out.println("//////////////////////////////////////////////////");
    System.out.println("NOTE: no map is loaded.");
    System.out.println("NOTE: default simulation parameters are loaded.");
    System.out.println("NOTE: no parameter changes allowed after a map load.");
    System.out.println("///////////////////////////////////////////////////");

    this.setupForm = setup;          // synch-up with the variables input form.

    this.loopRate = INITIAL_RATE;                // initialize sim update rate.
    this.movementHistory = false;  // reset the agent movement tracker control.
    this.initialized = false;                        // reset the control flag.

    Dimension screenSize = new Dimension ();
    screenSize = Toolkit.getDefaultToolkit().getScreenSize();   // display h-v.
    screenHeight = (int) screenSize.getHeight ();  // display screen dimension.
    screenWidth = (int) screenSize.getWidth ();    // display screen dimension.

    ///////////////////////////////////////////////////////////////////////////

    jfrm = new JFrame();                // set aside memory for the main frame.

    try
    { // select one of the three.
      UIManager.setLookAndFeel(
          "com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel"); // like a cloud.
    //    "com.sun.java.swing.plaf.windowsLookAndFeel");  // like a Windows UI.
    //    UIManager.getSystemLookAndFeelClassName());     // search internally.
      SwingUtilities.updateComponentTreeUI(jfrm);
    }
    catch (Exception e)
    {
      System.out.println("ERROR: unable to set look and feel.");
    }

    JMenuBar mbar = new JMenuBar();      // this is the menu bar in the JFrame.
    jfrm.setJMenuBar(mbar);           // load the menu bar into the main frame.

    ///////////////////////////////////////////////////////////////////////////

    jfrm.addComponentListener(this);    // enable the resizing tools interrupt.

    ///////////////////////////////////////////////////////////////////////////

    jfrm.addWindowListener ( new WindowAdapter ()
    {

      @Override
      public void windowClosing (WindowEvent event)
      {
        exitForm (event);
      }

    } );

    ///////////////////////////////////////////////////////////////////////////

    JMenu f = new JMenu("File");            // this is the overhead menu label.
    mbar.add(f);                                  // add the label to the menu.
    loadMap = new JMenuItem("Load Map");        // this is a label on a button.
    loadMap.addActionListener(this);           // the interrupt for the button.
    f.add(loadMap);                     // add the label to the drop-down menu.
    setupSim = new JMenuItem("Set Params");     // this is a label on a button.
    setupSim.addActionListener(this);  // the interrupt for button.
    f.add(setupSim);                    // add the label to the drop-down menu.
    f.add(new JSeparator());   // add this horizontal line separating sections.
    prepStats = new JMenuItem("Prep Stats");    // this is a label on a button.
    prepStats.addActionListener(this);             // the interrupt for button.
    f.add(prepStats);                   // add the label to the drop-down menu.
    f.add(new JSeparator());   // add this horizontal line separating sections.
    quitSim = new JMenuItem("Quit Sim");        // this is a label on a button.
    quitSim.addActionListener(new Terminate()); // add label to drop-down menu.
    f.add(quitSim);                     // add the label to the drop-down menu.

    ///////////////////////////////////////////////////////////////////////////

    JMenu c = new JMenu("Start/Stop");                                // ditto.
    mbar.add(c);
    startSim = new JMenuItem("Start Sim");
    startSim.addActionListener(this);
    c.add(startSim);
    stopSim = new JMenuItem("Stop Sim");
    stopSim.addActionListener(this);
    c.add(stopSim);
    continueSim = new JMenuItem("Continue");
    continueSim.addActionListener(this);
    c.add(continueSim);
    c.add(new JSeparator());
    stepSim = new JMenuItem("Step");
    stepSim.addActionListener(this);
    c.add(stepSim);
    c.add(new JSeparator());
    restartSim = new JMenuItem("Restart");
    restartSim.addActionListener(this);
    c.add(restartSim);

    ///////////////////////////////////////////////////////////////////////////

    JMenu uR = new JMenu("Screen Updates");                           // ditto.
    mbar.add(uR);
    fastUpdate = new JMenuItem("Daily");
    fastUpdate.addActionListener(this);
    uR.add(fastUpdate);
    mediumUpdate = new JMenuItem("Weekly");
    mediumUpdate.addActionListener(this);
    uR.add(mediumUpdate);
    slowUpdate = new JMenuItem("Monthly");
    slowUpdate.addActionListener(this);
    uR.add(slowUpdate);
    yearUpdate = new JMenuItem("Yearly");
    yearUpdate.addActionListener(this);
    uR.add(yearUpdate);

    ///////////////////////////////////////////////////////////////////////////

    JMenu rI = new JMenu("Resize Map");                               // ditto.
    mbar.add(rI);
    zoomIn = new JMenuItem("Zoom In");
    zoomIn.addActionListener(this);
    rI.add(zoomIn);
    zoomOut = new JMenuItem("Zoom Out");
    zoomOut.addActionListener(this);
    rI.add(zoomOut);

    ///////////////////////////////////////////////////////////////////////////

    JMenu t = new JMenu("History");                                   // ditto.
    mbar.add(t);
    historyOn = new JMenuItem("Tracking On");
    historyOn.addActionListener(this);
    t.add(historyOn);
    historyOff = new JMenuItem("Tracking Off");
    historyOff.addActionListener(this);
    t.add(historyOff);

    ///////////////////////////////////////////////////////////////////////////

    JMenu a = new JMenu("About");                                     // ditto.
    mbar.add(a);
    sayHelp = new JMenuItem("Help");
    sayHelp.addActionListener(new SayHelpPane());
    a.add(sayHelp);
    sayVersion = new JMenuItem("Version");
    sayVersion.addActionListener(new SayVersionPane());
    a.add(sayVersion);

    ///////////////////////////////////////////////////////////////////////////

    loadMap.setEnabled(true);
    setupSim.setEnabled(true);
    prepStats.setEnabled(false);
    historyOn.setEnabled(false);
    historyOff.setEnabled(false);
    zoomIn.setEnabled(false);
    zoomOut.setEnabled(false);
    yearUpdate.setEnabled(false);
    slowUpdate.setEnabled(false);
    mediumUpdate.setEnabled(false);
    fastUpdate.setEnabled(false);
    startSim.setEnabled(false);
    stopSim.setEnabled(false);
    continueSim.setEnabled(false);
    stepSim.setEnabled(false);
    restartSim.setEnabled(false);

    ///////////////////////////////////////////////////////////////////////////

    jfrm.setTitle ("Clans: Emergent, Related and Extended, Local-groups");
    jfrm.setSize(START_WINDOW_WIDTH, START_WINDOW_HEIGHT); //VGA size at start.
    jfrm.setResizable(false);          // disables the resizable frame feature.
    jfrm.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE); // 1 of 2 exits.
    jfrm.setLocation(50, 50);    // this is the upper left corner of the frame.
    jfrm.getContentPane().add(wPaper, BorderLayout.CENTER);   // add watermark.
    jfrm.setVisible(true);                          // make everything visible.
    jfrm.repaint();                // update the screen with the buffered data.

    ///////////////////////////////////////////////////////////////////////////

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method contains the User Interface.*/

  public void actionPerformed(ActionEvent userInterrupt)
  {

    Object userChoice = userInterrupt.getSource();          // convert by type.

    // --------------------- drop-down menu choice: Start/Stop ----------------
    if (userChoice == startSim)
    {
      if (this.initialized)
      {
        setUserStartRequest(true);                       // start the sim flag.
        setUserStopRequest (false);                // do not stop the sim flag.
        startSim.setEnabled(false);                    // control button light.
        stopSim.setEnabled(true);                                     // ditto.
        continueSim.setEnabled(false);                                // ditto.
        stepSim.setEnabled(false);                                    // ditto.
        restartSim.setEnabled(false);                                 // ditto.
        prepStats.setEnabled(false);                                  // ditto.
      }
    }
    else if (userChoice == stopSim)
    {
      setUserStartRequest(false);                      // stop simulating flag.
      startSim.setEnabled(false);                      // control button light.
      stopSim.setEnabled(false);                                      // ditto.
      continueSim.setEnabled(true);                                   // ditto.
      stepSim.setEnabled(true);                                       // ditto.
      restartSim.setEnabled(false);                                   // ditto.
      prepStats.setEnabled(true);                                     // ditto.
      System.out.println ("Simulation stopped by user.");
    }
    else if (userChoice == continueSim)
    {
      setUserStartRequest(true);                         // start the sim flag.
      startSim.setEnabled(false);                      // control button light.
      stopSim.setEnabled(true);                                       // ditto.
      continueSim.setEnabled(false);                                  // ditto.
      restartSim.setEnabled(false);                                   // ditto.
      prepStats.setEnabled(false);                                    // ditto.
      System.out.println ("Continue simulation.");
    }
    else if (userChoice == stepSim)
    {
      setUserStartRequest(true);                         // start the sim flag.
      wait (10);
      setUserStartRequest(false);                      // stop simulating flag.
      prepStats.setEnabled(false);                                    // ditto.
      System.out.println ("Simulation step requested.");
    }
    else if (userChoice == restartSim)
    {
      setRestart(true);                                // restart the sim flag.
      setUserStartRequest(true);                         // start the sim flag.
      restartSim.setEnabled(false);                    // control button light.
      setUserStartRequest(true);                         // start the sim flag.
      wait (10);
      setUserStartRequest(false);                      // stop simulating flag.
      startSim.setEnabled(true);                                      // ditto.
      stepSim.setEnabled(true);                                       // ditto.
      prepStats.setEnabled(false);                                    // ditto.

      System.out.println ("Simulation restarting with re-initialization.");
    }
    // --------------------- drop-down menu choice: Start/Stop ----------------

    // --------------------- drop-down menu choice: Screen Updates ------------
    else if (userChoice == yearUpdate)
    {
      setRepRate (YEARLY_UPDATE);                             // 364 day loop.
      yearUpdate.setEnabled(false);
      slowUpdate.setEnabled(true);
      mediumUpdate.setEnabled(true);
      fastUpdate.setEnabled(true);
    }
    else if (userChoice == slowUpdate)
    {
      setRepRate (MONTHLY_UPDATE);                              // 28 day loop.
      yearUpdate.setEnabled(true);
      slowUpdate.setEnabled(false);
      mediumUpdate.setEnabled(true);
      fastUpdate.setEnabled(true);
    }
    else if (userChoice == mediumUpdate)
    {
      setRepRate (WEEKLY_UPDATE);                                // 7 day loop.
      yearUpdate.setEnabled(true);
      slowUpdate.setEnabled(true);
      mediumUpdate.setEnabled(false);
      fastUpdate.setEnabled(true);
    }
    else if (userChoice == fastUpdate)
    {
      setRepRate (DAILY_UPDATE);                                 // 1 day loop.
      yearUpdate.setEnabled(true);
      slowUpdate.setEnabled(true);
      mediumUpdate.setEnabled(true);
      fastUpdate.setEnabled(false);
    }
    // --------------------- drop-down menu choice: Screen Updates ------------

    // --------------------- drop-down menu choice: Setup Sim -----------------
    else if (userChoice == setupSim)
    {
      if (!this.initialized)
      {
        System.out.println ("Parameter setup requested.");
        this.setupForm.parameterSetup();
      }
    }
    // --------------------- drop-down menu choice: Setup Sim -----------------

    // --------------------- drop-down menu choice: Quit ----------------------
    else if (userChoice == quitSim)
    {
      setUserStartRequest(false);                           // stop simulating.
      setUserStopRequest(true);                             // stop simulating.
      System.out.println ("Simulation ended normally.");
      System.exit(0);                                  // 2nd of 2 exit places.
    }
    // --------------------- drop-down menu choice: Quit ----------------------

    // --------------------- drop-down menu choice: Prepare Statistics --------
    else if (userChoice == prepStats)
    {
      prepStatisticsFiles = true;
      prepStats.setEnabled(false);                     // control button light.
    }
    // --------------------- drop-down menu choice: Prepare Statistics --------

    // --------------------- drop-down menu choice: Load DEM ------------------
    else if (userChoice == loadMap)
    {

      // guarantee a folder exists under the current runtime folder.
      String mapsPath = new String ("");
      try
      {
        File sample = new File(BIN_FOLDER + MAP_FOLDER);             // create.
        mapsPath = sample.getCanonicalPath();              // path to the maps.
        sample = null;
      }
      catch (IOException ex)
      {
        System.out.println("ERROR: GUI.open io exception.");
      }

      // this will open a pop-dialog window to retrieve the users filename.
      JFileChooser chooser = new JFileChooser(mapsPath);    // look here first.
      FileNameExtensionFilter filter = new     // prompt only these file types.
                      FileNameExtensionFilter( "*.dat; *.bmp", "dat", "bmp");
      chooser.setFileFilter(filter);            // apply the image type filter.
      int choice = chooser.showOpenDialog(jfrm);   // show the user dialog box.
      File fileName = chooser.getSelectedFile();   // receive the users choice.

      if (choice == JFileChooser.APPROVE_OPTION)
      {
        System.out.println("Map file loading in progress.");
        // --------------------------------------------------------------------
        // OK, we need the size of the map image in order to begin.
        BufferedImage bimg = null;       // a temporary object used for sizing.
        try
        {
          bimg = ImageIO.read(fileName);     // a throw-away read to get sizes.
        }
        catch (IOException ioe)
        {
          System.out.println("ERROR: GUI.actionPerformed getting bmp size");
        }

        this.horizontalLimit = bimg.getWidth();   // this value will be shared.
        this.verticalLimit = bimg.getHeight();    // this value will be shared.
        bimg = null;              // return the memory to the operating system.
        // --------------------------------------------------------------------
        // now that we know the size we can intake the image and manipulate it.
        this.bckGndImg = new BufferedImage //the bckgrnd display graphic panel.
                            (this.horizontalLimit, this.verticalLimit,
                             BufferedImage.TYPE_INT_ARGB);
        this.msImage = new BufferedImage                 // MS-BMP fomat image.
                          (this.horizontalLimit, this.verticalLimit,
                          BufferedImage.TYPE_3BYTE_BGR);
        this.imageArray = bmpToArray(fileName.toString()); //read in file data.
        this.imageArray = RGBToBGR(this.imageArray);  // convert array formats.
        this.imageArray = invertVerticalLines(this.imageArray);//BMP row rvrsl.
        // --------------------------------------------------------------------
        this.msImage = arrayToBffrdImage(this.imageArray);//cnvrt 2 BffrdImage.
        this.msImageScaled = arrayToBffrdImage(this.imageArray);      // ditto.
        this.imageArray = null;           // return memory to operating system.
        // --------------------------------------------------------------------
        this.bckGndImg = convertBffrdImg(this.msImage); // dupe & cnvrt bImage.
        scratchPad = new ImageIcon(this.bckGndImg);  // cnvrt Bffrd to Img Icn.
        scratchP = new JLabel(scratchPad);     // package Image Icon as Jlabel.
        // --------------------------------------------------------------------
        jfrm.getContentPane().remove(wPaper);   // remove watermark background.
        jfrm.setSize(START_WINDOW_WIDTH, START_WINDOW_HEIGHT); //size at start.
        jfrm.setResizable(true);                      // allow window resizing.
        jfrm.getContentPane().add(scratchP, BorderLayout.CENTER);   // ContPan.
        jfrm.setVisible(true);                      // make everything visible.
        jfrm.repaint();                // update the map display with new data.
        // --------------------------------------------------------------------
        // Four notes to keep in mind: 1) msImage is the original background
        // material, 2) bckGndImage is the user's working image, 3) the two
        // images are different types of BufferedImage, 4) although different
        // the two types are compatible.
        // --------------------------------------------------------------------
        setUserStartRequest(false);                    // stop simulating flag.
        loadMap.setEnabled(false);                     // control button light.
        setupSim.setEnabled(false);                                   // ditto.
        historyOn.setEnabled(true);                                   // ditto.
        historyOff.setEnabled(false);                                 // ditto.
        zoomIn.setEnabled(true);                                      // ditto.
        zoomOut.setEnabled(false);                                    // ditto.
        yearUpdate.setEnabled(true);                                  // ditto.
        slowUpdate.setEnabled(true);                                  // ditto.
        mediumUpdate.setEnabled(true);                                // ditto.
        fastUpdate.setEnabled(false);                                 // ditto.
        startSim.setEnabled(true);                                    // ditto.
        stopSim.setEnabled(false);                                    // ditto.
        continueSim.setEnabled(false);                                // ditto.
        stepSim.setEnabled(true);                                     // ditto.
        restartSim.setEnabled(false);                                 // ditto.
        // --------------------------------------------------------------------
        setupSim.setEnabled(false);    // disable further changes to variables.
        // --------------------------------------------------------------------
        this.initialized = true;                           // ready to proceed.

        System.out.println("GUI loading complete.");
      }
      else if (choice == JFileChooser.CANCEL_OPTION)
        System.gc();           // take this opportunity to collect the garbage.
    }
    // --------------------- drop-down menu choice: Load DEM ------------------

    // --------------------- drop-down menu choice: Resize Image --------------
    else if (userChoice == zoomIn)
    {

      System.out.println("Halted by zoom."); // attempt to stop the simulation.
      simRunning = getUserStartRequest();       // record the state of the sim.
      stopSim.setEnabled(false);                       // control button light.
      continueSim.setEnabled(false);                                  // ditto.
      stepSim.setEnabled(false);                                      // ditto.
      restartSim.setEnabled(false);                                   // ditto.
      this.userStartRequest = false;         // attempt to stop the simulation.
      wait (2000);                  // wait 2000 ms while the simulation loops.
      do {} while (this.simRunning); // loop awaiting confirmation sim stopped.

      if (imageZoomFactor <= IMAGE_ZOOM_MAXIMUM)
      {
        // rescale  working images - master (msImage) and target (bckGndImg).
        imageZoomFactor *= IMAGE_ZOOM_SIZE_STEP;
        this.msImageScaled = rescaledImage(BCKGND_IMAGE_TYPE, imageZoomFactor);
        this.bckGndImg = rescaledImage(MS_IMAGE_TYPE, imageZoomFactor);
        // --------------------------------------------------------------------
        scratchPad = new ImageIcon(this.bckGndImg);//instantiate the map image.
        scratchP = new JLabel(scratchPad);             // package map as label.
        scrollPane = new JScrollPane(scratchP); // put map inside the scroller.
        // --------------------------------------------------------------------
        jfrm.getContentPane().removeAll();                // remove background.
        jfrm.getContentPane().add(scrollPane);            // show finished map.
        jfrm.setVisible(true);                      // make everything visible.
        jfrm.repaint();            // update the screen with the buffered data.
      }

      if (imageZoomFactor < IMAGE_ZOOM_MAXIMUM &&
          imageZoomFactor > IMAGE_ZOOM_MINIMUM)
      {
        zoomIn.setEnabled(true);       // enlarge the depiction of the graphic.
        zoomOut.setEnabled(true);       // reduce the depiction of the graphic.
      }
      else
      {
        zoomIn.setEnabled(false);      // enlarge the depiction of the graphic.
        zoomOut.setEnabled(true);       // reduce the depiction of the graphic.
      }

      if (simRunning)
      {
        continueSim.setEnabled(true);                  // control button light.
        stepSim.setEnabled(true);                                     // ditto.
        restartSim.setEnabled(true);                                  // ditto.
      }
      else
      {
        startSim.setEnabled(true);                     // control button light.
        stepSim.setEnabled(true);                                     // ditto.
      }

    }
    else if (userChoice == zoomOut)
    {

      System.out.println("Halted by zoom."); // attempt to stop the simulation.
      simRunning = getUserStartRequest();       // record the state of the sim.
      stopSim.setEnabled(false);                       // control button light.
      continueSim.setEnabled(false);                                  // ditto.
      stepSim.setEnabled(false);                                      // ditto.
      restartSim.setEnabled(false);                                   // ditto.
      this.userStartRequest = false;         // attempt to stop the simulation.
      wait (2000);                  // wait 2000 ms while the simulation loops.
      do {} while (this.simRunning); // loop awaiting confirmation sim stopped.

      if (imageZoomFactor >= IMAGE_ZOOM_MINIMUM)
      {
        // rescale  working images - master (msImage) and target (bckGndImg).
        imageZoomFactor /= IMAGE_ZOOM_SIZE_STEP;
        this.msImageScaled = rescaledImage(BCKGND_IMAGE_TYPE,imageZoomFactor);
        this.bckGndImg = rescaledImage(MS_IMAGE_TYPE, imageZoomFactor);
        // --------------------------------------------------------------------
        scratchPad = new ImageIcon(this.bckGndImg);//instantiate the map image.
        scratchP = new JLabel(scratchPad);             // package map as label.
        scrollPane = new JScrollPane(scratchP); // put map inside the scroller.
        // --------------------------------------------------------------------
        jfrm.getContentPane().removeAll();                // remove background.
        jfrm.getContentPane().add(scrollPane);            // show finished map.
        jfrm.setVisible(true);                      // make everything visible.
        jfrm.repaint();            // update the screen with the buffered data.
     }

      if (imageZoomFactor < IMAGE_ZOOM_MAXIMUM &&
          imageZoomFactor > IMAGE_ZOOM_MINIMUM)
      {
        zoomIn.setEnabled(true);       // enlarge the depiction of the graphic.
        zoomOut.setEnabled(true);       // reduce the depiction of the graphic.
      }
      else
      {
        zoomIn.setEnabled(true);       // enlarge the depiction of the graphic.
        zoomOut.setEnabled(false);      // reduce the depiction of the graphic.
      }

      if (simRunning)
      {
        continueSim.setEnabled(true);                  // control button light.
        stepSim.setEnabled(true);                                     // ditto.
        restartSim.setEnabled(true);                                  // ditto.
      }
      else
      {
        startSim.setEnabled(true);                     // control button light.
        stepSim.setEnabled(true);                                     // ditto.
      }

    }
    // --------------------- drop-down menu choice: Resize Image --------------

    // --------------------- drop-down menu choice: Traces --------------------
    else if (userChoice == historyOn)
    {
      // cause map plotting methods to NOT over-write cohort movement traces.
      setTraces(true);
      historyOn.setEnabled(false);
      historyOff.setEnabled(true);
    }
    else if (userChoice == historyOff)
    {
      // cause map plotting methods to clear cohort movement history.
      setTraces(false);
      historyOn.setEnabled(true);
      historyOff.setEnabled(false);
    }
    // --------------------- drop-down menu choice: Traces --------------------

    System.gc();               // take this opportunity to collect the garbage.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a converted copy of the integer array data in the
      parameter.  A Buffered Image is the return.*/

  public BufferedImage arrayToBffrdImage(int anArray[])
  {

    BufferedImage anImage = new BufferedImage
                            (this.horizontalLimit, this.verticalLimit,
                             BufferedImage.TYPE_INT_ARGB);

    int pixel = 0;

    for (int v = 0; v < this.verticalLimit; v++)
    {
      for (int h = 0; h < this.horizontalLimit; h++ )
      {
        anImage.setRGB(h, v, anArray[pixel++]);
      }
    }

    return anImage;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method restores the GUI after the user choice to restart or not.*/

  public void awaken()
  {

    blankTheDisplay();
    if (imageZoomFactor == IMAGE_DEFAULT_SIZE)
    {
      zoomIn.setEnabled(true);         // enlarge the depiction of the graphic.
      zoomOut.setEnabled(false);        // reduce the depiction of the graphic.
    }
    else
    {
      if (imageZoomFactor < IMAGE_ZOOM_MAXIMUM &&
          imageZoomFactor > IMAGE_ZOOM_MINIMUM)
      {
        zoomIn.setEnabled(true);       // enlarge the depiction of the graphic.
        zoomOut.setEnabled(true);       // reduce the depiction of the graphic.
      }
      else
      {
        zoomIn.setEnabled(false);      // enlarge the depiction of the graphic.
        zoomOut.setEnabled(true);       // reduce the depiction of the graphic.
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method paints every pixel in the display map with the default.*/

  public void blankTheDisplay ()
  {

    int width = imageZoomFactor * this.horizontalLimit; // scale the dimension.
    int height = imageZoomFactor * this.verticalLimit;  // scale the dimension.

    for (int h = 0; h < width; h++)
    {
      for (int v = 0; v < height; v++)
      {
        this.bckGndImg.setRGB(h, v, this.msImageScaled.getRGB(h, v));
      }
    }

    jfrm.setVisible(true);                          // make everything visible.
    jfrm.repaint();                // update the screen with the buffered data.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an integer array of the converted the data in a
      BMP file whose name is fileName, whose size is m x n pixels, is in
      MS-BMP format, and is 24-bit color.*/

  public int[] bmpToArray(String fileName)
  {

    final int FILE_HEADER_LENGTH = 54;               // BMP file header length.

    int trash = 0;                                           // ignored status.
    int fileLen = getFileLength (fileName);       // get the input file length.
    int dataLen = fileLen - FILE_HEADER_LENGTH;  // data length without header.
    int outgoing[] = new int[dataLen];              // an array for the output.
    bmpWidth = 0;                                   // from header bytes 18-21.
    bmpHeight = 0;                                  // from header bytes 22-25.

    try
    {
      FileInputStream buffer = new FileInputStream(fileName); // open the file.
      for (int i = 0; i < FILE_HEADER_LENGTH; i++)    // throw away the header.
      {
        switch (i)
        {
          case 18: bmpWidth =  1 * buffer.read(); break;
          case 19: bmpWidth += 256 * buffer.read(); break;
          case 20: bmpWidth += 4096 * buffer.read(); break;
          case 21: bmpWidth += 65536 * buffer.read(); break;
          case 22: bmpHeight =  1 * buffer.read(); break;
          case 23: bmpHeight += 256 * buffer.read(); break;
          case 24: bmpHeight += 4096 * buffer.read(); break;
          case 25: bmpHeight += 65536 * buffer.read(); break;
          default: trash = buffer.read();
        }
      }
      for (int i = 0; i < dataLen; i++)                // capture all the data.
        outgoing[i] = buffer.read();  // a byte at a time ... move data across.
      buffer.close();                                        // close the file.
    }
    catch (IOException ioe)
    {
      System.err.println("Error in GUI.bmpToArray");
    }

    return outgoing;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is invoked when the component has been made invisible.*/

  public void componentHidden ( ComponentEvent e )
  {

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is invoked when the component position changes.*/

  public void componentMoved ( ComponentEvent e )
  {

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is invoked when the component's size changes.*/

  public void componentResized ( ComponentEvent e )
  {

    if ( e.getSource().equals ( jfrm ) )
    {

      if ( scrollPane == null )
        return;

      Component c = (Component) e.getSource();

      int newWidth = (int) c.getSize().getWidth();    // get the user's choice.
      int newHeight = (int) c.getSize().getHeight();  // get the user's choice.

      if (newWidth < screenWidth)  // so long as choice is LT screen size, OK.
        newWidth = Math.min ( newWidth, screenWidth);
      if (newHeight < screenHeight)
        newHeight = Math.min ( newHeight, screenHeight);

      // snap-back to fit the panel.
      jfrm.setSize(newWidth, newHeight); // maximum displayable screen surface.

    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is invoked when the component has been made visible.*/

  public void componentShown ( ComponentEvent e )
  {

    //add code here.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a Buffered Image of TYPE_3BYTE_BGR when provided
      with a Buffered Image of TYPE_INT_ARGB.*/

  public BufferedImage convertBffrdImg (BufferedImage imageA)
  {

    BufferedImage imageB = new BufferedImage (this.horizontalLimit,
                                              this.verticalLimit,
                                              BufferedImage.TYPE_3BYTE_BGR);

    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {
        imageB.setRGB(h, v, imageA.getRGB(h, v));
      }
    }

    return imageB;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method causes the form to close and the system to exit. */

  private void exitForm ( WindowEvent event )
  {

    int response = JOptionPane.showConfirmDialog (
                   jfrm,
                   "Terminate Clans?",
                   "WARNING - Data May Be Lost!",
                   JOptionPane.YES_NO_OPTION );

    switch ( response )
    {

      case JOptionPane.YES_OPTION:
        System.out.println ("Simulation ended normally.");
        System.exit ( 0 );
        break;                                          // choice made to quit.

      case JOptionPane.NO_OPTION:
        break;                                      // choice made to continue.

      case JOptionPane.CLOSED_OPTION:
        break;                           // closed without control choice made.

    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the length (in byte)s of the fileName provided.*/

  public static int getFileLength (String fileName)
  {

    int fileLength = 0;

   // find length of the file in bytes and number of integer records.
   try
    {
      RandomAccessFile fileIn = new RandomAccessFile(fileName, "r");
      fileLength = (int) fileIn.length();
      fileIn.close();
    }
    catch (IOException ioe)
    {
      System.err.println("Error: " + ioe);
    }

    return fileLength;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method uses the colors of the user's map as proxies for elevation.
      It returns the pointer passed to it representing the ecology AND within
      that ecology it will place terrain features according to map elevation.*/

  public Ecology getHeightMap (Ecology ecoSys)
  {

    RGB rgb = new RGB();                      // the tool set with RGB methods.
    int pixelColor = 0;                         // a encoded RGB integer value.
    int terrainFeature = 0;  // standard feature set like water, land, or road.

    for (int h = 0; h < bmpWidth; h++)
    {
      for (int v = 0; v < bmpHeight; v++)
      {
        pixelColor = this.msImage.getRGB (h, v);
        rgb.colorDisassembler(pixelColor);
        terrainFeature = terrainFeatureConverter (rgb.getRedComponent(),
                                                  rgb.getGreenComponent(),
                                                  rgb.getBlueComponent(),
                                                  h, v);
        ecoSys.setGridElevation (h, v, terrainFeature);
      }
    }
    rgb = null;                   // return the memory to the operating system.

    return ecoSys;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the vertical height of the currently loaded map.*/

  public int getMapHeight()
  {

    return bmpHeight;               //derived in bmpToArray, from header 22-25.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the horizontal width of the currently loaded map.*/

  public int getMapWidth()
  {

    return bmpWidth;                //derived in bmpToArray, from header 18-21.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the control flag associated with the user's choice
      to prepare data statistics files.*/

  public synchronized boolean getPrepStatistics()
  {

    boolean flag = this.prepStatisticsFiles;

    this.prepStatisticsFiles = false;

    return flag;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the repetition rate value.*/

  public long getRepRate()
  {

    return this.loopRate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the control flag stating user wants to restart.*/

  public synchronized boolean getRestart()
  {

    return this.doRestart;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the user start request flag.*/

  public synchronized boolean getUserStartRequest()
  {

    return this.userStartRequest;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the user stop request flag.*/

  public synchronized boolean getUserStopRequest()
  {

    return this.userStopRequest;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method inverts and array and returns it,  The inversion works
      line by line, along each vertical line, from top to bottom, moving
      horizontally left to right.*/

  public int[] invertVerticalLines(int source[])
  {

    final int H_MAX = bmpWidth;            // derived when the bmp was read-in.
    final int V_MAX = bmpHeight;

    final int LAST_CELL = H_MAX * V_MAX;
    int target[] = new int[LAST_CELL];

    int sourcePntr = 0;
    int sourceLimit = sourcePntr + H_MAX;
    int targetLimit = LAST_CELL;
    int targetPntr = targetLimit - H_MAX;
    do
    {
      targetPntr = targetLimit - H_MAX;
      targetLimit = targetPntr;
      sourceLimit = sourcePntr + H_MAX;
      do
      {
        target[targetPntr++] = source[sourcePntr++];
      }
      while (sourcePntr < sourceLimit);
    }
    while (sourcePntr < LAST_CELL);

    return target;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method plots to the display the data in a BufferedImage.*/

  public void plotTheDisplay (Ecology ecoSys)
  {

    String preyOption = "";               // the prey associable with the cell.
    float waterVolume = 0.0F;          // maximum water basin volume in a cell.
    int nrInCohort = 0;          // the number of hominid that are in the cell.
    int pixel = 0;      // a pixel marking where hominid(s) are in a grid cell.
    int width = imageZoomFactor * this.horizontalLimit; // scale the dimension.
    int height = imageZoomFactor * this.verticalLimit;  // scale the dimension.

    for (int h = 0; h < width; h++)
    {
      for (int v = 0; v < height; v++)
      {
        // --------------------------------------------------------------------
        if (run_time_display)
        {
          if (this.movementHistory)
            pixel = this.bckGndImg.getRGB (h, v);
          else
            pixel = this.msImageScaled.getRGB(h, v);

          // hominid in the cell.
          nrInCohort = ecoSys.getCohortCount (h/imageZoomFactor,
                                              v/imageZoomFactor);
          if (nrInCohort >= 1 ) pixel = RED;
        }
        else
        {

          if (foraging_display)
          {

            // prey in the cell.
            preyOption = ecoSys.getForageOption(h/imageZoomFactor,
                                                v/imageZoomFactor);
            if (preyOption.equals("fig"))
              pixel = ORANGE;
            else if (preyOption.equals("leaves"))
              pixel = GREEN;
            else if (preyOption.equals("monkey"))
              pixel = YELLOW;
            else if (preyOption.equals("termite"))
              pixel = RED;
            else if (preyOption.equals("tuber"))
              pixel = BROWN;
            else if (preyOption.equals("fish"))
              pixel = BLUE;
            else if (preyOption.equals("none"))
              pixel = BLACK;

          }
          else
          {

            // water basins in the cell.
            waterVolume = ecoSys.getBiomeWater (h/imageZoomFactor,
                                                v/imageZoomFactor);
            if (waterVolume >= 999.0F)
              pixel = BLUE;
            else if (waterVolume >= 4.4F)
              pixel = WHITE;
            else if (waterVolume >= 4.1F)
              pixel = GRAY_6;
            else if (waterVolume >= 3.6F)
              pixel = GRAY_5;
            else if (waterVolume >= 3.1F)
              pixel = GRAY_4;
            else if (waterVolume >= 2.6F)
              pixel = GRAY_3;
            else if (waterVolume >= 2.1F)
              pixel = GRAY_2;
            else if (waterVolume >= 1.6F)
              pixel = GRAY_1;
            else if (waterVolume >  0.9F)
              pixel = GRAY_0;
            else if (waterVolume == 0.0F)
              pixel = BLACK;

          }
        }
        // --------------------------------------------------------------------
        this.bckGndImg.setRGB (h, v, pixel);
        // --------------------------------------------------------------------
      }
    }

    jfrm.setVisible(true);                          // make everything visible.
    jfrm.repaint();                     // update the screen with the new data.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the default map scaled according to type & factor.*/

  public BufferedImage rescaledImage(int imageType, int factor)
  {

    int w = factor * this.msImage.getWidth();
    int h = factor * this.msImage.getHeight();

    BufferedImage targetImage = new BufferedImage (w, h, imageType);

    for (int y = 0; y < h; ++y)
    {
      for (int x = 0; x < w; ++x)
      {
        targetImage.setRGB ( x, y, this.msImage.getRGB (x/factor, y/factor) );
      }
    }

    return targetImage;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method converts an integer valued RGB array to an integer valued
      BGR color data array one third smaller.*/

  public int[] RGBToBGR(int array[])
  {

    int arrayLength = array.length;        // array with discrete integer data.
    int colorData[] = new int[arrayLength / 3];    // array for unified colors.
    RGB pixel = new RGB();                    // the tool set with RGB methods.

    int i = 0;
    int j = 0;
    boolean ok2Loop = true;
    do
    {

      colorData[i] = pixel.Hue(array[j + 2], array[j + 1], array[j + 0]);

        // avoid painful ArrayIndexOutOfBoundsExceptions!
      if ((j + 3) < arrayLength)
      {
        i = i + 1;
        j = j + 3;
      }
      else
        ok2Loop = false;

    } while (ok2Loop);

    return colorData;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the control flag associated with the user's choice
      to prepare data statistics files.*/

  public synchronized void setPrepStatistics(boolean newState)
  {

    this.prepStatisticsFiles = newState;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the repetition rate value.*/

  private void setRepRate(long newRate)
  {

    this.loopRate = newRate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the control flag stating user wants to restart.*/

  public synchronized void setRestart(boolean flag)
  {

    this.doRestart = flag;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method allows an external process to set-reset and internal
      flag that says the simulation is (or is not) running.*/

  public synchronized void setSimRunning(boolean flag)
  {

    this.simRunning = flag;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the status of the movement Traces feature.*/

  private void setTraces(boolean newState)
  {

    this.movementHistory = newState;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the user start request flag.*/

  private synchronized void setUserStartRequest(boolean newStatus)
  {

    this.userStartRequest = newStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the user stop request flag.*/

  private synchronized void setUserStopRequest(boolean newStatus)
  {

    this.userStopRequest = newStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method prepares the GUI during the user choice to restart or not.*/

  public void sleep()
  {

    prepStats.setEnabled(true);
    startSim.setEnabled(false);
    stopSim.setEnabled(false);
    restartSim.setEnabled(true);
    stepSim.setEnabled(false);
    zoomIn.setEnabled(false);
    zoomOut.setEnabled(false);
    setRestart(false);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns an integer that singularly identifies a pixel on the user
      map using its RGB components.  Currently, the pixel may be identified
      as either WATER, LO_LAND, or HI_LAND and can be mapped onto the
      terrain as elevation features.  This does not map the features.
      That must be done elsewhere.*/

  public int terrainFeatureConverter (int red, int green, int blue,
                                      int h, int v)
  {

    final int WATER_R =   0;              // rgb color integer value for pixel.
    final int WATER_G =   0;                                          // ditto.
    final int WATER_B = 255;                                          // ditto.

    final int LO_LAND_R =  34;             //rgb color integer value for pixel.
    final int LO_LAND_G = 177;                                        // ditto.
    final int LO_LAND_B =  76;                                        // ditto.

    final int HI_LAND_R = 128;             //rgb color integer value for pixel.
    final int HI_LAND_G = 128;                                        // ditto.
    final int HI_LAND_B =   0;                                        // ditto.

    if (red == WATER_R && green == WATER_G && blue == WATER_B)
      return water;

    else if (red == LO_LAND_R && green == LO_LAND_G && blue == LO_LAND_B)
      return lo_land;

    else if (red == HI_LAND_R && green == HI_LAND_G && blue == HI_LAND_B)
      return hi_land;

    else
    {
      System.out.println("ERROR MODE: GUI.terrainFeatureConverter data.");
      System.out.println("            h=" + h + " v=" + v);
      return 0; // ERROR MODE:
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method cause a momentary pause in ms in the executing thread.*/

  public static void wait (int delay)
  {

    try
    {
      Thread.sleep(delay);
    }
    catch (InterruptedException ie)
    {
      System.out.println ("ERROR: Clans.wait.");
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template()
  {

    //add code here.

  } // End method.
  //***************************************************************************

}
