//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    ArtificialGenetics.java. (Artificial Genetics)                   *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  This provides a basis for an artificial genetics.                *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                       // home directory.

import java.util.Random;                                      // Java language.

//
//*****************************************************************************
// Artificial Genetics

public class ArtificialGenetics
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  public final static int GENE_LENGTH = 8;   // hard coded count of characters.
  public final static int GENE_COUNT = 11;     // count of genes in chromosome.

  // Variables.
  private static int [] gaussDistribution = null;  // pointer to Gaussian data.
  private long rng_Seed = 0L;                   // this is the GLOBAL RNG seed.

  // Objects.
  public Random RNG = new Random();                 // an RNG object with seed.

  //***************************************************************************
  /** This is a default class constructor.*/

  public ArtificialGenetics()
  {

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This is the class constructor for Artificial Genetics.*/

  public ArtificialGenetics(long newSeed)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);    // instantiate an RNG object with seed.

    // during construction the agent Gauss Distribution data set will be built.
    gaussDistribution = makeGaussianDistribution();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method randomly assigns special characteristics to a chromosome.*/

  public String assignPhenotype()
  {

    final float MALE_SKEW = 0.495F;          // likelihood a hominid is a male.

    final int SEX_BIT = 7;                            // bit to which assigned.
    final int OTHER_NAMED_6 = 6;                      //                   etc.
    final int OTHER_NAMED_5 = 5;                      //                   etc.
    final int OTHER_NAMED_4 = 4;                      //                   etc.
    final int OTHER_NAMED_3 = 3;                      //                   etc.
    final int OTHER_NAMED_2 = 2;                      //                   etc.
    final int OTHER_NAMED_1 = 1;                      //                   etc.
    final int OTHER_NAMED_0 = 0;                      //                   etc.

    // remember bits in StringBuffers are counted from left to right as 0 - n.
    StringBuffer gene = new StringBuffer("00000000");       // create a buffer.
    // a hominids sex is set according to probability.
    Integer sex = null;
    if (RNG.nextFloat() < MALE_SKEW)                 // set this hominid's sex.
      sex = new Integer(1);                        // select either 0=F or 1=M.
    else
      sex = new Integer(0);                        // select either 0=F or 1=M.
    gene.deleteCharAt(SEX_BIT);                    // remove the dedicated bit.
    gene.insert(SEX_BIT, Integer.toBinaryString(sex));           // insert new.

    // randomly assign other_named_0 as another example of usage.
    //Integer dummy = new Integer(RNG.nextInt(2));
    //gene.deleteCharAt(OTHER_NAMED_0);
    //gene.insert(OTHER_NAMED_0, dummy.toBinaryString(dummy));

    return gene.toString();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method calculates the standard deviations among the values in the
      array.*/

  public double calculateStdDev (int array [], double mean)
  {

    double sigma = 0.0D;
    for (int counter = 0; counter < array.length; counter++)
      sigma = sigma + Math.pow( (((double) array [counter]) - mean) , 2);

    double variance = sigma / ((double) array.length);        // find variance.

    double stdDev = Math.sqrt(variance);        // take the standard deviation.

    return stdDev;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method takes an array parameter and returns the sum of the data.*/

  public int calculateSum(int [] array)
  {

    int sum = 0;

    for (int i = 0; i < array.length; i++)      // loop thru the group and sum.
    {
      array[i] = i;
      sum = sum + array[i];
    }

    return sum;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a value in the output range y of the Gauss function
      if a value x lies in the function domain and is the provided parameter.*/

  public int gaussianFunction(int x, double mean, double stdDev)
  {

    int result = 0;     // boundaries will need control due to rounding errors.
    if ((x >= 255) || (x <= 0))
      result = x;                                     // these are hard limits.
    else                             // all other values need to be calculated.
    {
      double sample = (double) x;    // convert integer sample to double value.

      double heightOfBell  = 21000.0D;  // these 4 are used to shape the curve.
      double widthOfBell = .55D;    // and are empirically-graphically defined.
      double bellFlanks = .39D;                                       // ditto.
      int    fineOffset = 2;                                          // ditto.

      double numerator = - Math.pow (((double) x - mean), 2);
      double denominator = 2.0D * Math.pow ((widthOfBell * stdDev), 2);
      double exponent = numerator / denominator;
      double exponential = Math.pow(Math.E, exponent);
      double fraction = 1.0D / (stdDev * (Math.sqrt(bellFlanks * Math.PI)));

      result = ((int)(heightOfBell * fraction * exponential)) - fineOffset;
    }

    return result;                                        // the final product.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a comma delimited version of the chromosome.*/

  public String getChromosomeString(String chromosome)
  {

    int geneValue = 0;    // temp var holds the gene currently being converted.
    String result = "";      // output String holding the delimited chromosome.

    for (int geneNr = 0; geneNr < ( GENE_COUNT - 1 ); geneNr++)
    {

      geneValue = getGeneValue ( chromosome, geneNr );
      result = result + new Integer ( geneValue ).toString () + ",";

    }

    geneValue = getGeneValue ( chromosome, ( GENE_COUNT - 1 ) );
    result = result + new Integer ( geneValue ).toString ();    // undelimited.

    return result;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the integer value of the gene, in the position
      requested, in chromosome provided.  Uses getGene to retreive the
      Allele sequence under examination.*/

  public int getGeneValue(String chromosome, int geneNumber)
  {

    return binStrToDec (getGene( chromosome, geneNumber), GENE_LENGTH);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the gene found in the position along the length of
      a chromosome. Genes are numbered from right to left with the smallest
      numbered position on the right and the smallest numbered position on
      left.  Genes are numbered n, n-1, n-2, ... 2, 1, 0  etc.*/

  public String getGene(String chromosome, int geneNumber)
  {

    int startAllele = (GENE_COUNT * GENE_LENGTH) -
                                    ((geneNumber * GENE_LENGTH) + GENE_LENGTH);
    int endAllele = startAllele + GENE_LENGTH;

    return chromosome.substring( startAllele, endAllele );

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method tests and returns the sex of the chromosome provided.*/

  public String getSex(String chromosome)
  {

    if (chromosome.charAt(chromosome.length() - 1) == '1')   //assume 1 = male.
      return "M";
    else
      return "F";

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method creates and returns one chromosome of GENECount length.
      A chromosome is an n-gene structure composed of concatenated "zeros"
      and "ones" (alleles).  The minimum length chromosome has one GENE.
      The maximum length (multi-gene) is undefined.*/

  public String makeChromosome()
  {

    final int RANDOM_TYPE = 0;          // default GENE type is uniform random.
    final int GAUSSIAN_TYPE = 1;       // value based on Gaussian distribution.

    String chromosome = new String();                            // chromosome.

    for (int i = 0; i < (GENE_COUNT - 1); i++)
      chromosome = chromosome + makeGene(GAUSSIAN_TYPE);

    chromosome = chromosome + assignPhenotype(); //append the phenotype scheme.

    // DIAGNOSTIC
    //  System.out.println ("ArtificialGenetics.makeChromosome gene: " +
    //                      getGeneValue (chromosome, 3));
    // DIAGNOSTIC

    return chromosome;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the construction and return of an array of 256,
      8-bit, integer encoded, Gaussian Normal distributed values.  When
      returned, the array data is in bell curve configuration.*/

  public int [] makeGaussianDistribution()
  {

    int geneMaxValue = (int) (Math.pow(2, GENE_LENGTH));           // calc 2^n.
    int array [] = new int[geneMaxValue];       // array that holds everything.

    for (int i = 0; i < geneMaxValue; i++)
      array[i] = signedByte(i);                       // integers -128 to +127.

    double sum = (double) calculateSum (array); // sum the values in the array.
    double mean = sum / ((double) geneMaxValue);         // calculate the mean.
    double stdDev = calculateStdDev (array, mean);   // take the Std Deviation.

    for (int i = 0; i < geneMaxValue; i++)
      if (i == (geneMaxValue / 2))
        array[i] = geneMaxValue - 1;
      else if (i == (geneMaxValue - 1))
        array[i] = 0;
      else
        array[i] = gaussianFunction(array[i], mean, stdDev);      // calculate.

      // DIAGNOSTIC
      // for (int j = 0; j < geneMaxValue; j++)
      // System.out.println("j: " + j + ", " + array[j]);
      // System.exit(0);

    return array;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method creates and returns one n-character GENE string composed
      of concatenated "zero" and "one" characters (GENEs). The parameter,
      randomType, determines if the GENE returned is made from the uniform
      random pool of integer values generated by the RNG or is based on a
      Gaussian distribution.  Gaussian normal is default.*/

  public String makeGene(int randomType)
  {

    String rawResult = "";                    // the result as initially built.
    int geneMaxValue = 0;          // the largest value a gene may be plus one.
    geneMaxValue = (int) (Math.pow(2, GENE_LENGTH));          // calculate 2^n.
    Integer gene = null;                              // a pointer to the gene.
    int limitCheck = 0;                          // an error checking variable.

    switch (randomType)
    {
      case 1:   // randomly select a value from a Gaussian normal distribution.
        limitCheck = gaussDistribution[RNG.nextInt(geneMaxValue)];
        gene = new Integer(limitCheck);
      break;
      default:    // randomly select a gene from a uniform random distribution.
        limitCheck = RNG.nextInt(geneMaxValue);
        gene = new Integer(limitCheck);
      break;
    }

    rawResult = Integer.toBinaryString(gene);       // turn gene into a string.
    int rawResultLen = rawResult.length();     // the length of the raw result.
    int neededPadding = GENE_LENGTH - rawResultLen;     // calc needed padding.

    String outBuffer = "";                    // make an all zeros padded gene.
    for (int i = 0; i < neededPadding; i++)
      outBuffer = outBuffer + "0";

    outBuffer = outBuffer.concat(rawResult);
    return outBuffer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method can induce a random mutation in the chromosome provided.  A
      constant MUTATION_PROBABILITY determines if the mutation occurs.*/

  public String randomMutation(String chromosome)
  {

    // per Nachman, 2000 should be 2.5 X 10 ^ -8 per nucleotide per generation.
    // final int MUTATION_PROBABILITY = 5;         // using equivalent factors.
    // final int SEARCH_SPACE = 200000000;           // search space for above.

    final int MUTATION_PROBABILITY = 3;    // using 100K equivalent and 0.003%.
    final int SEARCH_SPACE = 100000;                 // search space for above.

    int chromosomeLen = chromosome.length();
    StringBuffer candidate = new StringBuffer(chromosome); // mutatn candidate.

    int mutationNumber = RNG.nextInt(SEARCH_SPACE);     // out of many samples.
    if (mutationNumber < MUTATION_PROBABILITY)     // check if mutation occurs.
    {
      int mutationSite = RNG.nextInt(chromosomeLen);    // pick an allele site.
      Integer mutant = new Integer(RNG.nextInt(2));    // select either 0 or 1.
      candidate.deleteCharAt(mutationSite);           // remove the buffer bit.
      candidate.insert(mutationSite, Integer.toBinaryString(mutant)); //mutate.
    }

    return candidate.toString();                  // may or may not be mutated.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a new chromosome "child" as a result of the process
      of genetic cross-over.  The child chromosome is the same length as its
      parents.  Alleles from the parents are site-wise selected, at random,
      to construct the child.  Finally, the child chromosome undergoes
      random mutation.*/

  public String reproduce(String parent_1, String parent_2)
  {

    String child = new String();                               // working copy.
    StringBuffer p_1 = new StringBuffer(parent_1);             // working copy.
    StringBuffer p_2 = new StringBuffer(parent_2);             // working copy.

    for (int i = 0; i < p_1.length(); i++)
    {
      if (RNG.nextBoolean())                 // select one parent or the other.
        child = child + p_1.charAt(i);      // append the new genetic material.
      else
        child = child + p_2.charAt(i);      // append the new genetic material.
    }

    child = randomMutation(child);    // send child to undergo random mutation.

    return child;             // return the newly constructed child chromosome.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This routine returns a decimal version of Binary Number converted into
      decimal.  The function assumes the binary string to be a unsigned
      value.  Format is msb....lsb.  Conversion for values greater than
      24-bits is undefined.  Exponent is the working bit width of the number
      to be converted.*/

  public int binStrToDec (String binary, int exponent)
  {

    int result = 0;                                      // the working result.
    int tmpExp = exponent;          // the minimum output word bit-width width.
    int i = 0;                                                    // A counter.

    do
    {

      --tmpExp;

      if (binary.charAt(i) == '1')
        result += Math.pow (2, tmpExp);

    } while (++i < exponent);

    return result;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an 8-bit signed number between -128 and +127
      (inclusive) when provided with an integer valued 0 to 255.  Parameter
      values greater than 255 or smaller than 0 are truncated.  Examples
      are in=0 out=-128, in=127 out=-1, in=128 out=0, and in=255 out=127.*/

  public int signedByte (int number)
  {

    int dummy = 0;

    if (number > 255) number = 255;                              // hard limit.
    if (number < 0) number = 0;                                  // hard limit.

    if ((number >= 128) && (number <= 255))
      dummy = number - 128;

    if ((number >=   0) && (number <= 127))
      dummy = - Math.abs(number - 128);

    return dummy ;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template()
  {

    //add code here.

  } // End method.
  //***************************************************************************

}
