//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Socioecology.java. (Social ecology)                              *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  This emulates kin-related properties and mechanisms.             *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in Java to model proto-humans.                           *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                       // home directory.

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;

import root.ModelConstraints;                          // program dependencies.

//
//*****************************************************************************
// Socioecology

public class Socioecology extends ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  private final double W_DAYS = 1.00D;// Westermarck-days prohibit incest prct.

  // Variables.
  private long rng_Seed;                        // this is the GLOBAL RNG seed.

  private int mDistance = 1;           // Murdock (1949) geneological distance.
  public boolean isAlpha = false;           // the individual selected to lead.

  // Objects.
  public  Hominid self = null;        // pointer to the self available to self.
  private Hominid mother = null;            // a pointer to the mother of self.
  private Hominid father = null;            // a pointer to the father of self.
  private ArrayList <Hominid> siblings = null;             // list of siblings.
  private ArrayList <Hominid> children = null;             // list of children.
  private Hominid mate = null; // a pointer to an opposite sex partner of self.

  private Hashtable <Hominid, Integer> localGroup = null; // the others nearby.

  private static Random RNG = null;          // pointer to standard RNG object.

  //***************************************************************************
  /** This is a class constructor.  It makes adults between 15-35 years old.*/

  public Socioecology (long newSeed, Hominid thisSelf)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);                // an RNG object with seed.

    this.self = thisSelf;        // complete the self-reference pointer system.

    this.siblings = new ArrayList <Hominid> ();            // list of siblings.
    this.children = new ArrayList <Hominid> ();            // list of children.

    // the localGroup is a superset of all others the self has encountered.
    this.localGroup = new Hashtable <Hominid, Integer> ();// with contact days.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method adds a child to the self's list of children placing other
      at the beginning of the list.  This ordering keeps the youngest child
      in the list with the lowest numbered index.*/

  public void addChild (Hominid other)
  {

    this.children.add (0, other);       // insert at the beginning of the list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING:  This is a long method.

      This method scans the entire population and controls Alpha presence.
      Only one Alpha is allowed per cohort.  Further, it is possible an
      Alpha may arrive (N) needing reassignment or (D) dead.  Both cases
      make the selection of a new local band Alpha imparative.  The
      local cohort is supplied by parameter.*/

  // DEPRECATED AND NEEDS REWORK
  public void alphaCheck (ArrayList <Hominid> cohort, Hominid hominid)
  {

    ArrayList <Hominid> band = null; // another name for this hominid's cohort.

    // count up all local-band homis who claim to be viable Alphas.
    int alphaCount = 0;      // assume, intially, there is no local band Alpha.
    int hNumber = 0;               // this is a hominid's number in the cohort.
    do                        // find out how many local-band Alphas there are.
    {

      hominid = cohort.get(hNumber);              // get a copy of the hominid.
      if (hominid.bio.getHominidLifeStatus().equals("D"))
        ;  // ignore (D) dead and those in migration.
      else
        if (hominid.kin.getIsAlpha()) alphaCount++;//record the Alpha presence.

    } while (++hNumber < cohort.size());      // work across the entire cohort.

    // since exactly one local-band Alpha must exist per cohort, any other
    // number is a problem that must be cleared up.
    if (alphaCount != 1)
    {

      // first, either last Epoch one or more Alphas joined the cohort and
      // now there are too many Alphas or all local-group homis are (D) or (N)
      // and are not viable. ie., alphaCount = 0.  Thus, from among what
      // ever (D), (N), (J), or (A) local-band remnants that are present,
      // suppress all claims to Alpha.
      hNumber = 0;                           // points to a particular hominid.
      do
      {
        hominid = cohort.get(hNumber);            // get a copy of the hominid.
        if (hominid.bio.getHominidLifeStatus().equals("D"))
          ; //ignore (D) dead and those in migration.
        else                                           // suppress Alpha claim.
          if ( hominid.kin.getIsAlpha() ) hominid.kin.setIsAlpha(false);
      } while (++hNumber < cohort.size());    // work across the entire cohort.

      // second, if there exists a viable local-band Alpha candidate then
      // the cohort must be split into two groups (local-band vs. all others)
      // so attempt to split the group if split is possible at all.
      band = new ArrayList <Hominid> ();      // this will hold the local band.
      ArrayList <Integer> pullOutsList= null; // special members of the cohort.
      pullOutsList= new ArrayList <Integer> ();  // and pointers into that set.
      hNumber = 0;                                // points to a particular HG.
      do
      {
        hominid = cohort.get(hNumber);            // get a copy of the hominid.
        if (hominid.bio.getHominidLifeStatus().equals("D"))
          ; //ignore (D) dead and those in migration.
        else
        {
          band.add(hominid);  // add this homi to the band for Alpha selection.
          pullOutsList.add(hNumber);     // note their placement in the cohort.
        }
      } while (++hNumber < cohort.size());    // work across the entire cohort.

      // if a split was possible, ie., there exists at least one viable Alpha
      // candidate, then select the Alpha, reassemble the cohort, and exit.
      if (band.size() != 0)
      {
        makeAlpha (band);   // go away and select one Alpha for the local band.

        int pBand = 0;         // just a pointer into the band, agent by agent.
        hNumber = 0;                            //  pointer into the HG cohort.
        do         // reintegrate the band (and its new Alpha) into the cohort.
        {
          hominid = band.get(pBand);       // get a copy of the (next) hominid.
          hNumber = pullOutsList.get(pBand);//get their position in the cohort.
          cohort.set(hNumber, hominid);   // put the homi back into the cohort.
        } while (++pBand < band.size());// work across the members of the band.
      }
      band = null;                           // give the memory back to the OS.
      pullOutsList = null;                   // give the memory back to the OS.

    }  // end if test for one local-band Alpha.
    else
    ; //System.out.println("in LifeExperiences.dailyLife(alphaCheck)");//DIAG.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method appends a littermate onto the end of the self's list of
      siblings placing the new littermate at the end of the list.  This
      ordering preserves the youngest child in the list with the lowest
      numbered index.*/

  public void appendSibling (Hominid litterMate)
  {

    this.siblings.add (litterMate);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the count of self's children.*/

  public int childrenCount ()
  {

    return this.children.size();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method clears the children list.*/

  private void clearChildren ()
  {

    this.children.clear ();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method clears the siblings list.*/

  private void clearSiblings ()
  {

    this.siblings.clear ();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is the argument other a blood kin
      according to a Murdoch constellation extent equal to one (1).  In other
      words is self the sibling of, the male or female biological parent of,
      or the child of the argument other.*/

  public boolean consanguinalKin (Hominid other)
  {

    return isSiblingOf (other) || isParentOf (other) || isOffspringOf (other);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method attempts to find the other in the self's local group list.
      Others in that list are those who having been in close physical contact
      with self.  Close contact is defined as being in the same grid cell.
      at the end of an epoch.*/

  public int contactDays (Hominid other)
  {

    Integer nrOfDaysKnown = 0;

    if (this.localGroup.containsKey (other))
      nrOfDaysKnown = this.localGroup.get (other);

    return nrOfDaysKnown.intValue ();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the momentary average of the contact days this
      hominid has had with the members of its local group excluding its mom.*/

  public double contactDayAvg ()
  {

    int members = 0;                           // count of local group members.
    int sum = 0;                                    // sum of all contact days.

    for (Hominid homi : getLocalGroup ())
    {

      if (homi.getAgentID() != getMother().getAgentID())
      {
        members++;
        sum += contactDays (homi);
      }

    }

    return (double) sum / (double) members;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the last child added to the self's list of children.
      The child is NOT guaranteed to be a suckling.  Use isSuckling() or
      hasSuckling() to see if child is pre-weaning aged and thus a suckling.
      Warning - always check hasChildren() before calling this function.*/

  public Hominid getLastChild ()
  {

    if (!hasChildren ())
    {

      System.out.println ("ERROR: Socioecology.getChild () req children" +
                          " pointer w/o" +
                          " first checking for count of existing children.");
      System.exit (0);

    }

    return this.children.get (0);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the i'th child in the self's list of children.
      Warning - always check hasChildren() before calling this function.*/

  public Hominid getChild (int i)
  {

    if (!hasChildren ())
    {

      System.out.println ("ERROR: Socioecology.getChild (i) req children" +
                          " pointer w/o" +
                          " first checking for count of existing children.");
      System.exit (0);

    }

    return this.children.get (i);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns this Agent's children.*/

  public ArrayList <Hominid> getChildren ()
  {

    return this.children;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the self's dad pointer.*/

  public Hominid getFather ()
  {

    return this.father;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the Hominid's alpha status.*/

  public boolean getIsAlpha ()
  {

    return this.isAlpha;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns an ArrayList of the member of self's local group held
      in the Hashtable data structure localGroup.*/

  public ArrayList <Hominid> getLocalGroup ()
  {

    ArrayList <Hominid> list = new ArrayList <Hominid> ();

    if (this.localGroup.size () > 0)
    {

      Enumeration <Hominid> enumeratedList = this.localGroup.keys ();
      while( enumeratedList.hasMoreElements())
      {
        list.add (enumeratedList.nextElement());
      }

    }

    return list;     // NOTE: this can return a list of 0 members but not null.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the self's opposite sex partner.*/

  public Hominid getMate ()
  {

    return this.mate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the self's mom pointer.*/

  public Hominid getMother ()
  {

    return this.mother;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns self's Murdock kin geneological constellation distance.*/

  public int getMurdockDist ()
  {

    return this.mDistance;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns this Agent's siblings list.*/

  public ArrayList <Hominid> getSiblings ()
  {

    return this.siblings;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question does self have one or more children.*/

  public boolean hasChildren ()
  {

    if (this.children.size() > 0)
      return true;
    else
      return false;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question does self have any non-weaned children.
      Implicitly, if self has no children then all are weaned.  Else, if there
      is one that is not (weaned) return that Boolean.*/

  public boolean hasSuckling ()
  {

    boolean atLeastOneIsNursing = false;    // assume all childeren are weaned.

    // first check if self has any children?
    if (hasChildren ())
    {

      // there is at least one child (dead or alive).
      for (Hominid child: this.children)
      {

        // is there (are there) any children of self under the weaning age?
        if (child.kin.isSuckling ())
          atLeastOneIsNursing = true;              // at least one is unweaned.

      }

    }

    return atLeastOneIsNursing;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates (on a cohort basis) the individual group list(s)
      of all the agents identified in the localCohort parameter.*/

  public void informGroupsOf (ArrayList <Hominid> localCohort)
  {

    Hominid member = null;
    Hominid hominid = null;

    int limit = localCohort.size ();
    for (int i = 0; i < limit; i++)
    {

      member = localCohort.get (i);
      for (int j = 0; j < limit; j++)
      {
        hominid = localCohort.get (j);
        if (hominid != member)
          updateGroup (member);    // FILTER:  only deltas allowed.
                    // updateGroup(m) operates on the individual self.
      }

    }

    member = null;                // return the memory to the operating system.
    hominid = null;               // return the memory to the operating system.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method operates from the perspective of mom and informs each of
      mom's children that they have a new sibling.  Before it is finished,
      it updates the sibling lists of all of mom's children.  It is assumed
      newLitterMate was inserted into mom's list of children before this
      was called else the results will be invalid.  This tests for extant
      litter mates before any agent updates are made.  The youngest child
      in the liss has the lowest numbered indice.*/

  public void informSiblingsOf (Hominid newLitterMate)
  {

    Hominid litterMate = null;                     // pointer to a litter mate.

    // OPERATES: on current litter mates
    // use mom's children list to identify the existing siblings.  Start
    // at the end of the list and move to the front, skipping the first
    // one because the first one is the newest member (ie newLitterMate).
    // The purpose is to add the newLitterMate to list of each existing
    // litter mate.
    for (int i = (this.children.size () - 1); i > 0; i--)
    {

      litterMate = this.children.get (i);   // checks only littermates in list.
      if (litterMate.getExtant())         // excludes dead (dying) littermates.
        litterMate.kin.insertSibling (newLitterMate);

    }

    // OPERATES: on newest litter mate
    // use mom's children list to identify the existing siblings.  Since
    // the newest litter mate is the first in mom's list, skip it, start
    // with the second one, and move to the end.  The purpose is to add
    // each of the existing litter mates to the newLitterMate's list.
    for (int i = 1; i < this.children.size (); i++)
    {

      // add the other litter mates to the newest litter mate's list.
      litterMate = this.children.get (i);
      if (litterMate.getExtant())
        this.children.get (0).kin.appendSibling (litterMate);

    }

    litterMate = null;            // return the memory to the operating system.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method inserts the littermate into the self's list of siblings
      placing the new littermate at the beginning of the list.  This
      ordering keeps the youngest child in the list with the lowest
      numbered index.*/

  public void insertSibling (Hominid litterMate)
  {

    this.siblings.add (0, litterMate);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self a daughter.  It assumes self
      has parents.  Equivalent to asking for confirmation self is female.*/

  public boolean isDaughter ()
  {

    return isFemale();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self female and the biological
      offspring of the argument other.*/

  public boolean isDaughterOf (Hominid other)
  {

    boolean answer = false;        // assume self is not the daughter of other.

    if (isDaughter ())
    {

      if ( other.getAgentID() == this.mother.getAgentID() ||
           other.getAgentID() == this.father.getAgentID() )
      {
        answer = true; // other's ID matches either my mother or my fathers ID.
      }

    }

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the answer to the question is this Hominid a father.*/

  public boolean isFather ()
  {

    if (isMale())
      return hasChildren ();
    else
      return false;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the answer to the question, is self the biological
      father of the argument other.*/

  public boolean isFatherOf ( Hominid other )
  {

    boolean answer = false;                   // assume self is not the father.

    if (isFather())
    {

      // Uterine kin rules prohibit the inclusion of paternity insight.
      // Self does not know if they are a biological father or not.  Thus,
      // we must ask of (other) who is "their" biological father in order to
      // get sufficient reference to answer the question, isFatherOf(other).
      if (this.self.getAgentID() == other.getFatherID())
        answer = true;           // is my ID equivalent to other's father's ID.

    }

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self a female.*/

  public boolean isFemale ()
  {

    return this.self.bio.getSex ().equals("F");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self a male.*/

  public boolean isMale ()
  {

    return this.self.bio.getSex ().equals("M");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the answer to the question is this Hominid a mother.*/

  public boolean isMother ()
  {

    if (isFemale ())
      return hasChildren ();
    else
      return false;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the answer to the question, is self the biological
      mother of the argument other.*/

  public boolean isMotherOf ( Hominid other )
  {

    boolean answer = false;                   // assume self is not the mother.

    if (isMother())
    {

      if (this.self.getAgentID() == other.getMotherID())
        answer = true;           // is my ID equivalent to other's mother's ID.

    }

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the answer to the question, is other within the
      bounds of self's Murdock kinship limits according to the constellation
      distance limit provided in the parameter.*/

// add code here.
  public boolean isMurdockKin ( Hominid other, int distance )
  {

    ArrayList <Hominid> constellation = new ArrayList <Hominid> ();

    // add siblings
    // add parents (mother & father)
    // add offspring
    // extend focus to parents (mother & father)
    // extent focus to each offspring
    // add mother's constellation (do not add those already contained)
    // add father's constellation (ditto)
    // add each offspring's constellation (ditto)

    boolean answer = false;
    int selfKinLimit = this.mDistance;

// add code here.

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self the biological offspring of,
      i.e, is the son or the daughter of, the argument other.*/

  public boolean isOffspringOf (Hominid other)
  {

    return isDaughterOf (other) || isSonOf (other);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self the biological parent of
      the argument other.*/

  public boolean isParentOf (Hominid other)
  {

    return (isMotherOf (other) || isFatherOf (other));

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the answer to the question is the self a sibling.*/

  public boolean isSibling ()
  {

    if (this.siblings.size () > 0)
      return true;
    else
      return false;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the answer to the question, do any of the IDs that
      self associates with its siblings match the argument other's ID.  In
      other words is other known to me from my list of siblings?*/

  public boolean isSiblingOf (Hominid other)
  {

    boolean answer = false;                     // assume self has no siblings.

    if (isSibling())
    {

      // choose to access the list sequentially, not randomly.
      for (Hominid sibling : getSiblings())
      {

        if (sibling.getAgentID() == other.getAgentID())
        {
          answer = true;    // other's ID was among a list of my sibling's IDs.
        }

      }

    }

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self a son.  It assumes self
      has parents.  Equivalent to asking for confirmation self is male.*/

  public boolean isSon ()
  {

    return isMale();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self male and the biological
      offspring of the argument other.*/

  public boolean isSonOf (Hominid other)
  {

    boolean answer = false;             // assume self is not the son of other.

    if (isSon ())
    {

      if ( other.getAgentID() == this.mother.getAgentID() ||
           other.getAgentID() == this.father.getAgentID() )
      {
        answer = true; // other's ID matches either my mother or my fathers ID.
      }

    }

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is self a non-weaned infant.*/

  public boolean isSuckling ()
  {

    return this.self.bio.getAge() < this.self.bio.getWeaningAge();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is a long method.

      This method inpects a list of unknown size and returns an array of the
      list members with one chosen to be group Alpha.  Initially, there is a
      preference for selecting a male over a female troop member as Alpha.
      Moreover, there is a preference to selecting an individual above a
      certain age.  Both of these preferences are overridden as troop
      (membership) constituency dictates reassesment of preference.  Assume
      the source list size is G.T. zero, i.e., that there exists a candidate
      pool over which to search for a suitably aged and gendered agent.*/

  // DEPRECATED AND NEEDS REWORK
  public ArrayList <Hominid> makeAlpha (ArrayList <Hominid> band)
  {

    final float GENDER_BIAS = 0.01F;   // smaller number reflects greater bias.
    final int AGE_MIN = 19; // initial minimum choice for the age of the Alpha.

    String gender = "";
    if (RNG.nextFloat() > GENDER_BIAS)       // rendomly select initial gender.
      gender = "M";
    else
      gender = "F";

    // now that its been decided what the troop gender-preferenced Alpha will
    // be, check if there exists a minimum-aged, gender-preferred homi in the
    // cohort suitable for Alpha installment.  If there is no one suitable
    // then switch sexes. Since this is both sex and age keyed, the next
    // step is to lower the age requirement on the primary sex choice and
    // to repeat until an Alpha can be found and marked.
    int preferredAge = AGE_MIN;         // the minimum age to start the search.
    int tries = 1;           // this will be the first attempt of perhaps many.
    int nrAvailable =0; // initially no suitable alpha candidates can be found.
    while (nrAvailable == 0)       // acknowledging the initial sex preference.
    {
      tries = 1;        // this will be the first attempt of perhaps many more.
      do
      {
      //nrAvailable = countBySex(band, gender, preferredAge);    // head count.
        if (nrAvailable == 0)    // no minimum-aged, gender-preferred HG found.
        {
          if (gender == "M")                     // swap the gender preference.
            gender = "F";
          else
            gender = "M";
          tries++;                                                // try again.
        }
        //  you're "done" after either 1 or 2 tries, there is no 3.
      } while ((tries < 3) && (nrAvailable == 0));
      if (nrAvailable == 0) preferredAge--; //no gender-preferred so lower age.
    }          // you now have an Alpha candidates list with one or more names.

    // make a list of the indices of the  Alpha eligibles.
    ArrayList <Integer> alphas = new ArrayList <Integer> ();
                      //alphas = makeListByType (band, gender, preferredAge);

    // since we know there must exist at least one non-incoming cohort member
    // (else we wouldn't be in this method), select the first non-incoming
    // hominid for Alpha status.
    Hominid alpha = null;               // create a pointer to the Alpha agent.
    int alphaIndex = 0;                          // the index to the new Alpha.
//  do
//  {
      alphaIndex = alphas.get(RNG.nextInt(alphas.size())); //get random index.
      alpha = band.get(alphaIndex);           // get a copy of the Alpha agent.
//  } while (alpha.getMigration());   // do not allow non-local HG to be Alpha.

    alpha.kin.setIsAlpha(true);                   // make the copy a new Alpha.
    band.set(alphaIndex, alpha);      // replace the original HG with the copy.

    return band;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method removes a child from the self's list of children.*/

  private void removeChild (int i)
  {

    this.children.remove (i);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method removes a member from the self's list of in group members.*/

  private void removeGroupMember (Hominid other)
  {

    this.localGroup.remove (other);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method removes a member from the self's list of siblings.*/

  private void removeSibling (int i)
  {

    this.siblings.remove (i);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the self's entire list of children.*/

  private void setChildren (ArrayList <Hominid> kids)
  {

    this.children = kids;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the self's geneological "dad" pointer.  It SHOULD
      ONLY BE ACCESSED AT BIRTH. Use addChild() to extend the list of
      self's offspring.*/

  public void setFather (Hominid dad)
  {

    this.father = dad;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the Hominid's alpha status.*/

  public void setIsAlpha (boolean newStatus)
  {

    this.isAlpha = newStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This set the self's opposite sex partner.*/

  public void setMate (Hominid affine)
  {

    this.mate = affine;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the self's geneological "mom" pointer.  It SHOULD
      ONLY BE ACCESSED AT BIRTH. Use addChild() to extend the list of
      self's offspring.*/

  public void setMother (Hominid mom)
  {

    this.mother = mom;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets self's Murdock kin geneological constellation distance.*/

  public void setMurdockDist ( int newLimit )
  {

    this.mDistance = newLimit;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the self's entire list of siblings.*/

  private void setSiblings (ArrayList <Hominid> others)
  {

    this.siblings = others;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates self's localGroup data structure by making sure
      other is in the group and that the number of self-other days has been
      updated.  Others not found in the localGroup will be added and they
      will get 1-day self-other contact credit.  For others already in the
      group, their self-other contact time will be incremented by 1*/

  public void updateGroup (Hominid other)
  {

    int daysOfContact = contactDays (other);

    Object o = null;

    // NOTE: a member already in the group will NOT EQUAL ZERO DAYS CONTACT.
    if (daysOfContact == 0)
    {
      o = this.localGroup.put (other, new Integer (1));
    }
    else
    {
      o = this.localGroup.replace (other, new Integer (++daysOfContact));
    }

    o = null;                   // return the memory to the operating system.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates self's awareness of the membership of the local
      cohort.  */

  public void updateGroup (ArrayList <Hominid> localCohort)
  {

    for (Hominid member : localCohort)
    {

      // don't add yourself to your own local group list.
      if (member.getAgentID() != this.self.getAgentID())
      {
        updateGroup (member);
      }

    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question is the argument "other" a blood kin
      of self according to a Murdoch constellation extent equal to one (1)
      but without explicit knowledge of the most senior male member.  In
      other words is self the sibling of, the female parent of, or a child
      with only materlineal knowledge of the argument other.*/

  public boolean uterineKin (Hominid other)
  {

    boolean answer = false;   // assume other is not part of the constellation.

    // FIRST, determine if other is female.
    if (other.kin.isFemale())
    {

      // NEXT, do self and other recognize each other by direct reference
      // to shared matrilineal connection.
      if ( other.getAgentID() == this.mother.getAgentID() )
      {
        answer = true;                    // other's ID matches my mother's ID.
      }
      // CROSS_CHECK
      if ( other.kin.isMotherOf (this.self) )
      {
        answer = true;                          //other claims to be my mother.
      }

    }

    // LAST, combine the foregoing answer with just two more questions.
    answer = isSiblingOf (other) || isMotherOf (other) || answer;

    return answer;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method answers the question has other been in the acquantanceship
      of, and in close proximity to, self such that, disregarding any
      pre-existing sexual relationship between self and other, mating between
      self and other would otherwise be proscribed by the Westermarck Effect.*/

  public boolean westermarckEffect (Hominid other, int epoch)
  {

    boolean wEffect = false;    // no, the Westermarck effect is NOT indicated.

    int daysInContact = contactDays (other);  // nr days in contact with other.
    if (daysInContact > 0)
    {
      // Westermarck days (W_DAYS) are an arbitrary percentage of the
      // momentary average number of days this agent has known all members
      // of its (past and present) local group.  The ratio of daysInContact
      // to the product of W_DAYS and the averageContactDays number is the
      // deciding factor.  Mating is not recommended if the daysInContact
      // are more than the computed Westermarck days value.  W_DAYS are an
      // arbitrary value.  The entire Westermarck function is a heuristic.
      if ((double) daysInContact > ((double) W_DAYS * contactDayAvg ()))
        wEffect = true;                                     // disallow mating.
    }

    return wEffect;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}