//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      This file.                                                       *
//* CLASS    Setup.java. (Setup - a support Class)                            *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  To be a startup-time toolset that sets-up runtime parameters.    *
//* METHOD   By means of GUI software prompt the user for their input.        *
//* DETAILS  Swing forms.                                                     *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.tools;                                   // sub-directory storage.

import java.awt.BorderLayout;                                 // Java language.
import java.awt.Color;                                        //         ditto.
import javax.swing.JFrame;                                    //         ditto.
import javax.swing.JOptionPane;                               //         ditto.
import javax.swing.JPanel;                                    //         ditto.
import javax.swing.SwingUtilities;                            //         ditto.
import javax.swing.JTabbedPane;                               //         ditto.
import javax.swing.UIManager;                                 //         ditto.
import javax.swing.UnsupportedLookAndFeelException;           //         ditto.

import java.util.ArrayList;                                   //         ditto.
import java.util.Random;                                      //         ditto.

//
//*****************************************************************************
// Setup

public class Setup
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  final String RNG_SWITCH_STR = "/rngSource=";              // the switch name.
  final String TRUE_RNG = "TrueRNG";                     // a console argument.
  final String FACTORY_RNG = "factoryRNG";               // a console argument.

  // Variables.
  private boolean okToExit = false;    // monitors user choic to exit the form.
  private boolean setupFormVisible = false; // monitors the status of the form.
  private boolean setupFormInUse = false;   // monitors the status of the form.

  // These variables are GLOBALLY visible and set here as default values.  All
  // of them are intended to be reset/set by the user during Setup Time.

  // MODEL "MODES & DETERMINISM"
  // --------------------------------------------------------------------------
  // Installation, operating system, and hardware RNG dependent USB dongle.
  String TRUE_RNG_PORT = "COM5";  // if using TrueRNG then must use COM port 5.
  long CALIB_SEED = 0;            // alternatively a fixed seed can be defined.

  // NOTE: RUN_TIME must be set FALSE in order to use the two DIAGNOSTIC modes.
  boolean RUN_TIME_DISPLAY = true;              //true=normal false=DIAGNOSTIC.
  boolean FORAGING_DISPLAY = true;              //true=forage false=waterBasin.
  // --------------------------------------------------------------------------
  // MODEL "MODES & DETERMINISM"

  // MODEL "HOMINID POPULATION(S)"
  // --------------------------------------------------------------------------
  int MAXIMUM_GROUP_SIZE = 1;                          // max nr in seed group.
  int NUMBER_OF_GROUPS = 1;                        // max nr of hominid groups.

  boolean FIXED_GRID = false; // if T, seed-group fixed grid must be >= 4Km sq.
  boolean USE_ALTRUISM = true;               // can engage altruistic behavior.
  // --------------------------------------------------------------------------
  // MODEL "HOMINID POPULATION(S)"

  // MODEL "ECOLOGICAL CONSTRAINTS"
  // --------------------------------------------------------------------------
  float RATE_OF_LOWLAND_WATER_BASINS = 0.5F;       // prcnt cells water basins.
  float RATE_OF_HIGHLAND_WATER_BASINS = 0.3F;      // prcnt cells water basins.

  float MIN_WATER = 1.0F;         // minimum gals of water each basin will get.
  float RANDOM_WATER = 3.50F;          // some basins get up to this much more.

  float INITIAL_PRODUCTIVITY = 99.0F;              // initial prcnt production.
  float FORAGING_COST = 1.0F;              // prcnt cost to terrain if foraged.
  float FORAGING_RCVRY_RATE = 0.50F;               // prcnt terrain rcvry rate.

  float FLORA_SUCCESS_BIAS = 0.50F;              // min seasonal foragable prd.
  float FAUNA_SUCCESS_BIAS = 0.50F;              // min seasonal foragable prd.
  // --------------------------------------------------------------------------
  // MODEL "ECOLOGICAL CONSTRAINTS"

  // MODEL "CIRCUMSCRIPTION"
  // --------------------------------------------------------------------------
  int WATER_DAYS = 2;               // max allowed consecutive epochs in water.

  float CHANCE_OF_PREDATION_HI = 0.0000140F;             // dy, 0.50% yr, esti.
  float CHANCE_OF_PREDATION_LO = 0.0000001F;             // dy, 0.04% yr, Will.

  float INFANTICIDE_MORT = 0.000055F;                    // daily, 2.0% annual.
  float DISEASE_THRESHOLD = 0.000055F;                   // daily, 2.0% annual.
  // --------------------------------------------------------------------------
  // MODEL "CIRCUMSCRIPTION"

  // MODEL "HOMINID PHYSICAL CONSTRAINTS"
  // --------------------------------------------------------------------------
  // equivalent to eating 100% of a 2700 kCal diet composed entirely of tubers.
  float DAILY_MEALS = 3.48F;               // "multiplier" best case situation.

  int FEAR_STEP = 2;                 // n X 10 meter movement distance per day.
  int WATER_STEP = 15;               // n X 10 meter movement distance per day.
  int FOOD_STEP = 15;                // n X 10 meter movement distance per day.
  int GROUP_STEP = 30;               // n X 10 meter movement distance per day.

  float WATER_START_AMOUNT = 0.7F;               // water amount to start with.
  float WATER_CARRYING_LIMIT = 1.0F;              // only hand-to-mouth (gals).

  float FOOD_START_AMOUNT = 0.7F;                 // food amount to start with.
  float FOOD_CARRYING_LIMIT_F = 0.7F;              // only hand to mouth (lbs).
  float FOOD_CARRYING_LIMIT_M = 0.9F;              // only hand to mouth (lbs).
  // --------------------------------------------------------------------------
  // MODEL "HOMINID PHYSICAL CONSTRAINTS"

  // MODEL "HOMINID COGNITIVE TRAITS"
  // --------------------------------------------------------------------------
  int PREY_ATTITUDE_CHANGE_THRESHOLD = 0364;                 //  1  day = 0001.
  //  1 week equiv = 0007.
  //  1 year equiv = 0364.
  //  5 year equiv = 1820.
  // 10 year equiv = 3640.
  // 15 year equiv = 5460.

  int GROUPED_DRIVES = 3;   // nr of executed highest priorty Maslowian drives.
  // --------------------------------------------------------------------------
  // MODEL "HOMINID COGNITIVE TRAITS"

  // Objects.
  private JFrame jfrm = null;            // the overall container for the pane.
  private JTabbedPane setupPane = null; // the container for the tabbed panels.
  private JPanel circumscription = null;  // a tabbed pane for data collection.
  private JPanel ecology = null;          // a tabbed pane for data collection.
  private JPanel hominids = null;         // a tabbed pane for data collection.
  private JPanel modes = null;            // a tabbed pane for data collection.

  //***************************************************************************
  /** This is the class constructor.*/

  public Setup ()
  {

    // add code here as needed.

  }
  //***************************************************************************

  //***************************************************************************
  /** This method closes then removes from memory the user choice form.*/

  public void closeSetupForm()
  {

    if (this.setupFormVisible)
    {
      this.jfrm.setVisible(false);               // attempt to remove the form.
    }

    this.jfrm = null;             // return the memory to the operating system.
    this.setupPane = null;        // return the memory to the operating system.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public void createCircumscriptionPanel()

  {

    circumscription = new JPanel();
    circumscription.setLayout(null);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public void createEcologyPanel()

  {

    ecology = new JPanel();
    ecology.setLayout(null);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public void createHominidsPanel()
  {

    hominids = new JPanel();
    hominids.setLayout(null);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public void createModesPanel()
  {

    modes = new JPanel();
    modes.setLayout(null);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public long getCALIB_SEED ()
  {

     return this.CALIB_SEED;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getCHANCE_OF_PREDATION_HI()
  {

    return CHANCE_OF_PREDATION_HI;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getCHANCE_OF_PREDATION_LO()
  {

    return CHANCE_OF_PREDATION_LO;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getDAILY_MEALS ()
  {

     return DAILY_MEALS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getDISEASE_THRESHOLD ()
  {

    return DISEASE_THRESHOLD;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFAUNA_SUCCESS_BIAS ()
  {

    return FAUNA_SUCCESS_BIAS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getFEAR_STEP ()
  {

    return FEAR_STEP;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public boolean getFIXED_GRID ()
  {

    return FIXED_GRID;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFLORA_SUCCESS_BIAS ()
  {

    return FLORA_SUCCESS_BIAS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFOOD_CARRYING_LIMIT_F ()
  {

    return FOOD_CARRYING_LIMIT_F;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFOOD_CARRYING_LIMIT_M ()
  {

    return FOOD_CARRYING_LIMIT_M;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFOOD_START_AMOUNT ()
  {

    return FOOD_START_AMOUNT;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getFOOD_STEP ()
  {

    return FOOD_STEP;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFORAGING_COST ()
  {

    return FORAGING_COST;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public boolean getFORAGING_DISPLAY ()
  {

    return FORAGING_DISPLAY;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getFORAGING_RCVRY_RATE ()
  {

    return FORAGING_RCVRY_RATE;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getGROUPED_DRIVES ()
  {

    return GROUPED_DRIVES;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getGROUP_STEP ()
  {

    return GROUP_STEP;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getINFANTICIDE_MORT ()
  {

    return INFANTICIDE_MORT;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getINITIAL_PRODUCTIVITY ()
  {

    return INITIAL_PRODUCTIVITY;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getMAXIMUM_GROUP_SIZE ()
  {

    return MAXIMUM_GROUP_SIZE;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getMIN_WATER ()
  {

    return MIN_WATER;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getNUMBER_OF_GROUPS ()
  {

    return NUMBER_OF_GROUPS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getPREY_ATTITUDE_CHANGE_THRESHOLD ()
  {

    return PREY_ATTITUDE_CHANGE_THRESHOLD;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getRANDOM_WATER ()
  {

    return RANDOM_WATER;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getRATE_OF_HIGHLAND_WATER_BASINS ()
  {

    return RATE_OF_HIGHLAND_WATER_BASINS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getRATE_OF_LOWLAND_WATER_BASINS ()
  {

    return RATE_OF_LOWLAND_WATER_BASINS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public boolean getRUN_TIME_DISPLAY ()
  {

    return RUN_TIME_DISPLAY;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public String getTRUE_RNG_PORT ()
  {

    return TRUE_RNG_PORT;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public boolean getUSE_ALTRUISM ()
  {

    return USE_ALTRUISM;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getWATER_CARRYING_LIMIT ()
  {

    return WATER_CARRYING_LIMIT;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getWATER_DAYS ()
  {

    return WATER_DAYS;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public float getWATER_START_AMOUNT ()
  {

    return WATER_START_AMOUNT;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the users value choice.*/

  public int getWATER_STEP ()
  {

    return WATER_STEP;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /**This method setups and controls the tabbed user input parameter form.*/

  public void parameterSetup()
  {

    final String PANEL_TITLE = "Clans: Parameter Setup";       // custom named.
    final int PANEL_WIDTH = 400;                               // custom sized.
    final int PANEL_HEIGHT = 500;                              // custom sized.

    this.jfrm = new JFrame();            // the overall container for the pane.
    this.setupPane = new JTabbedPane(); // the container for the tabbed panels.

    jfrm.setTitle(PANEL_TITLE);
    jfrm.setSize(PANEL_WIDTH, PANEL_HEIGHT);
    jfrm.setBackground(Color.gray);

    JPanel topPanel = new JPanel();
    topPanel.setLayout(new BorderLayout());
    jfrm.getContentPane().add(topPanel);

    // Create the tab pages
    createModesPanel();                   // a tabbed pane for data collection.
    createHominidsPanel();                // a tabbed pane for data collection.
    createEcologyPanel();                 // a tabbed pane for data collection.
    createCircumscriptionPanel();         // a tabbed pane for data collection.

    setupPane = new JTabbedPane();
    setupPane.addTab("Modes & Determinism", modes);
    setupPane.addTab("Hominid Population(s)", hominids);
    setupPane.addTab("Ecological Constraints", ecology);
    setupPane.addTab("Circumscription", circumscription);
    topPanel.add(setupPane, BorderLayout.CENTER);

    try
    {
      UIManager.setLookAndFeel(
       "com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");    // like a cloud.
       //"com.sun.java.swing.plaf.windowsLookAndFeel");   // like a Windows UI.
       //UIManager.getSystemLookAndFeelClassName());      // search internally.
       SwingUtilities.updateComponentTreeUI(jfrm);
    }
    catch (ClassNotFoundException | InstantiationException |
           IllegalAccessException | UnsupportedLookAndFeelException e)
    {
      System.out.println("ERROR: unable to set look and feel.");
    }

    jfrm.setVisible(true);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method parses the console command line for a valueSought. If the
      string value can be found then, it will be returned. Else, it returns
      the empty string and a failure to find error message.*/

  public String parseLine(String testLine, String valueSought)
  {

    int pntr = 0;                                 // a pointer into the string.
    String argument = "";                              // a temporary variable.

    if (testLine.contains(valueSought))         // check if valueSought exists.
    {

      pntr = testLine.indexOf(valueSought);    // where the valueSought starts.
      argument = testLine.substring(pntr);                   // find a section.
      pntr = pntr + valueSought.length(); // find first cell after valueSought.
      argument = argument.substring(pntr);                   // find a section.
      argument = argument.trim();

    }
    else
    {
      System.out.println("ERROR: command line argument not found.");
    }

     return argument;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a Boolean value that answers the question:
      "usefixed?" This is equivalent to asking does the user want to use the
      Java factory Class (TRUE) as the random number generator source or
      the TrueRNG USB dongle (FALSE). The default is the with the
      factoryClass.*/

  public boolean readConsole(String[] userSwitch)
  {

    String TrueRNG = TRUE_RNG;                           // a console argument.
    String factoryRNG = FACTORY_RNG;                     // a console argument.
    String lineArgument = "";                 // the user's line, empty or not.
    boolean decision = true;                   // default is the factory Class.
    String tempVar = "default";     // a String RNG source notice for the user.

    try
    {

      lineArgument = parseLine(userSwitch[0], RNG_SWITCH_STR);
      if (lineArgument.equals(TrueRNG))
      {
        decision = false;
        tempVar = TrueRNG;
        System.out.println("NOTICE: Success reading user RNG choice.");
      }
      else
      {
        if (lineArgument.equals(factoryRNG))
        {
          decision = true;
          tempVar = factoryRNG;
          System.out.println("NOTICE: Success reading user RNG choice.");
        }
        else
        {
          decision = true;
          tempVar = "default";
          System.out.println("Please check the console RNG choice.");
        }
      }
    }
    catch (ArrayIndexOutOfBoundsException aioobe)
    {
      decision = true;
      tempVar = "default";
      System.out.println("NOTICE: No valid RNG choice detected. ");
    }
    finally
    {
      System.out.println("NOTICE: Random number source will be the " +
                         tempVar + ".");
    }

    return decision;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method cause a momentary pause in ms in the executing thread.*/

  public void wait(int delay)
  {

    try
    {
      Thread.sleep(delay);
    }
    catch (InterruptedException ie)
    {
      System.out.println("ERROR: Clans.wait.");
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template()
  {

    return false;

  } // End method.
  //***************************************************************************

}
